//
//  Storage.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 9/15/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import CoreData
import UIKit
import SwiftyJSON

class CentralStorage {
    
    let HASH_KEY = "hashString"
    
    let managedObjectContext: NSManagedObjectContext
    var delegate: NSFetchedResultsControllerDelegate?
    let local : LocalStorage
    let sync: SyncQueue
    var mapping: DataMapping
    var entityName: String
    
    init(moc: NSManagedObjectContext, delegate: NSFetchedResultsControllerDelegate?, entityName: String){
        self.managedObjectContext = moc
        self.delegate = delegate
        self.entityName = entityName
        
        self.local = LocalStorage(managedObjectContext: managedObjectContext)
        self.sync = SyncQueue.getInstance(managedObjectContext)
        self.mapping = DataMapping(localStorage: local)
    }
    
    // MARK: - GENERAL
    func synchronise(){
        sync.processQueue()
    }
    
    func getMapping()->NSDictionary?{
        return nil
    }
    
    func getOutputMapping()->NSDictionary?{
        return getMapping()
    }
    
    func getPostMapping()->NSDictionary?{
        return getOutputMapping()
    }
    
    func getRelationsURLMapping()->NSDictionary?{
        return nil
    }
    
    func getRelationsAttrMapping()->NSDictionary?{
        return nil
    }
    
    func getManagedObjectForURL(url: NSURL) -> NSManagedObject{
        let appDelegate = UIApplication.sharedApplication().delegate as! AppDelegate
        let objectId = appDelegate.getPersistentStoreCoordinator().managedObjectIDForURIRepresentation(url)
        return managedObjectContext.objectWithID(objectId!)
    }
    
    func dateForJSON(date: NSDate = NSDate())->String{
        let dateFormatter = NSDateFormatter()
        dateFormatter.dateFormat = "yyyy-MM-dd HH:mm:ss"
        return dateFormatter.stringFromDate(date)
    }
    
    func prepareRequestData(event: Event)->SyncRequestData?{
        var item: NSManagedObject? = nil
        if(event.itemPointer != nil){
            item = getManagedObjectForURL(event.itemPointer!)
        }
        switch event.method{
        //case "GET":
            //return prepareGetRequestData(event, item: item)
        case "POST":
            return preparePostRequestData(event, item: item)
        case "PUT":
            return preparePutRequestData(event, item: item)
        case "DELETE":
            return prepareDeleteRequestData(event, item: item)
        default:
            NSLog("No request data generator for method %@", event.method)
            return nil
        }
    }
    
    func requestSuccessful(request: SyncRequestData!, response: NSHTTPURLResponse!, data: AnyObject?){
        NSLog("requestSuccessful %@", request.url)
        if data != nil {
            NSLog("vysledek z fronty je %@", makeJSONString(data))
        }
        
        var item: NSManagedObject? = nil
        
        if(request.event.itemPointer != nil){
            item = getManagedObjectForURL(request.event.itemPointer!)
        }
        
        switch request.event.method{
        //case "GET":
           // getRequestSuccessful(request, response: response, data: data, item: item)
        case "POST":
            handlePostResponse(request, response: response, data: data, item: item)
        case "PUT":
            handlePutResponse(request, response: response, data: data, item: item)
        case "DELETE":
            handleDeleteResponse(request, response: response, data: data, item: item)
        default:
            NSLog("No response handler for method %@", request.event.method)
        }
    }
    
    internal func getStorageClassName(eventType: String)->String{
        return eventType + "Storage"
    }
    
    // MARK: - GET
    func get(indexPath:NSIndexPath)->NSManagedObject?{
        return local.get(indexPath)
    }
    
    func getByKey(urlParts: [String], index: Int)->NSManagedObject?{
        if(index >= urlParts.count){
            return nil
        }else{
            return getByName(urlParts[index])
        }
    }
    
    func getByName(name: String?)->NSManagedObject?{
        if(name != nil){
            let predicate  = NSPredicate(format: "name = %@", name!)
            let frc = local.getAll(entityName, predicate: predicate, sort: getDefaultSort(), delegate: nil)
            if(frc.fetchedObjects?.count>0){
                return frc.fetchedObjects![0] as? NSManagedObject
            }
        }
        return nil
    }
    
    /**
     Fecthes the first object with given title. The entity must have column named title of type String.
     
     - parameter title: The searched title
     
     - returns: The fetched object
     */
    func getByTitle(title: String)->NSManagedObject?{
        let predicate = NSPredicate(format: "("+getDefaultPredicate()+") AND (title = %@)", title)
        let objects = local.getAll(entityName, predicate: predicate, sort: getDefaultSort(), delegate: nil)
        if(objects.fetchedObjects != nil && objects.fetchedObjects?.count > 0){
            return objects.fetchedObjects?[0] as? NSManagedObject
        }else{
            return nil
        }
    }
    
    func getDefaultPredicate()->String{
        return " (hidden = 0 OR hidden = nil) "
    }
    
    func getDefaultSort()->[NSSortDescriptor]{
        // Must be implemented in subclasses!
        abstract()
        return [NSSortDescriptor()]
    }
    
    func getAll()->NSFetchedResultsController{
        let predicate = NSPredicate(format: getDefaultPredicate())
        return local.getAll(entityName, predicate: predicate, sort: getDefaultSort(), delegate: delegate)
    }
    
    func clearCache(){
        local.clearCache()
    }
    
    func getItemURL(item: NSManagedObject)->String{
        abstract()
        return ""
    }

    func getItemByURL(url: String)->NSManagedObject?{
        let sanitized = sanitizeURL(url)
        let parts = (sanitized as NSString).pathComponents
        return getItemByURLParts(parts)
    }
    
    func getItemByURLParts(urlParts: [String])->NSManagedObject?{
        abstract()
        return nil
    }

    // MARK: - CREATE
    func newEntity() -> NSManagedObject {
        return local.newEntity(entityName)
    }
    
    func create(item: NSManagedObject){
        local.create(item)
        if(Settings.syncEnabled){
            sync.createEvent("POST", handlerClass: getStorageClassName(entityName), itemId: item.objectID.URIRepresentation())
            //sync.processQueue()
        }
    }
    
    func getCreateURL(item: NSManagedObject)->String{
        abstract()
        return ""
    }
    
    func preparePostRequestData(event: Event, item: NSManagedObject?)->SyncRequestData?{
        let url = getCreateURL(item!)
        let data = mapping.getRequestJSON((item as! Model), mapping: getPostMapping()!)
        return SyncRequestData(event: event, url: url, bodyString: data)
    }
    
    func handlePostResponse(request: SyncRequestData!, response: NSHTTPURLResponse!, data: AnyObject?, item: NSManagedObject?){
        if(response.statusCode == 201){ // created
            if(item != nil && data != nil){
                mapping.updateLocal(item!, json: data!, mapping: getMapping()!)
            }
        }else if(response.statusCode == 409){ // conflict (unique key violation) -> update local by remote values
            if(item != nil && data != nil){
                mapping.updateLocal(item!, json: data!, mapping: getMapping()!)
            }
        }else{
            sync.wrongResponse()
        }
    }
    
    // MARK: - UPDATE
    func update(item: NSManagedObject){
        if(item.managedObjectContext != nil){
            local.update(item)
            if(Settings.syncEnabled){
                sync.createEvent("PUT", handlerClass: getStorageClassName(entityName), itemId: item.objectID.URIRepresentation())
                //sync.processQueue()
            }
        }
    }
    
    func preparePutRequestData(event: Event, item: NSManagedObject?)->SyncRequestData?{
        let url = getItemURL(item!)
        let data = mapping.getRequestJSON((item as! Model), mapping: getOutputMapping()!)
        return SyncRequestData(event: event, url: url, bodyString: data)
    }
    
    func handlePutResponse(request: SyncRequestData!, response: NSHTTPURLResponse!, data: AnyObject?, item: NSManagedObject?){
        if(item != nil){
            if(response.statusCode == 202){ // accepted
                if(data != nil){
                    mapping.updateLocal(item!, json: data!, mapping: getMapping()!)
                }
            }else if(response.statusCode == 404){ // not found -> delete local item
                local.delete(item!)
            }else if(response.statusCode == 409){ // conflict -> update local by remote values
                if(data != nil){
                    mapping.updateLocal(item!, json: data!, mapping: getMapping()!)
                }
            }else{
                sync.wrongResponse()
            }
        }
    }
    
    // MARK: - DELETE
    func delete(item: NSManagedObject){
        if(item.managedObjectContext != nil){
            if(Settings.syncEnabled){
                local.hide(item)
                sync.createEvent("DELETE", handlerClass: getStorageClassName(entityName), itemId: item.objectID.URIRepresentation())
                //sync.processQueue()
            }else{
                local.delete(item)
            }
        }
    }
    
    func prepareDeleteRequestData(event: Event, item: NSManagedObject?)->SyncRequestData?{
        let url = getItemURL(item!)
        var data = ["type": entityName]
        if let type = (item as! Model).getEventType(){
            data["type"] = type
        }
        if let hash = item!.valueForKey(HASH_KEY) as! String?{
            data["hash"] = hash
        }
        return SyncRequestData(event: event, url: url, bodyString: makeJSONString(data))
    }
    
    func handleDeleteResponse(request: SyncRequestData!, response: NSHTTPURLResponse!, data: AnyObject?, item: NSManagedObject?){
        if(item != nil){
            if(response.statusCode == 204){ // no content -> delete local item
                NSLog("mazu lokalni")
                local.delete(item!)
            }else if(response.statusCode == 404){ // item already deleted on server
                    local.delete(item!)
            }else if(response.statusCode == 409){ // conflict => return local item and update it by server values
                local.unhide(item!)
                if(data != nil){
                    mapping.updateLocal(item!, json: data!, mapping: getMapping()!)
                }
            }else{
                sync.wrongResponse()
            }
        }
    }
    
    // MARK: - INCOMING EVENTS
    
    func handleIncomingPostEvent(event: JSON){
        let item = mapping.createLocal(event["data"], entityName: entityName, mapping: getMapping()!)
        mapping.saveRelationsByURL(item, url: event["uri"].stringValue, mapping: getRelationsURLMapping())
        mapping.saveRelationsByAttr(item, receivedJSON: event["data"], mapping: getRelationsAttrMapping())
    }
    
    func handleIncomingPutEvent(event: JSON){
        let item = getItemByURL(event["uri"].stringValue)
        if(item != nil){
            mapping.updateLocal(item!, json: event["data"].rawValue, mapping: getMapping()!)
            // nema smysl -- mapping.saveRelationsByURL(item!, url: event["uri"].stringValue, mapping: storage!.getRelationsURLMapping())
        }else{
            // maybe create a new one??
        }
    }
    
    func handleIncomingDeleteEvent(event: JSON){
        let item = getItemByURL(event["uri"].stringValue)
        if(item != nil){
            local.delete(item!)
        }
    }
}
