//
//  DataMapping.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 10/13/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import CoreData
import SwiftyJSON

class DataMapping{
    
    let RELATIONAL_ENTITY_DELIMITER = "."
    let arrayAttributes: [String] = []
    let jsonStringAttributes: [String] = ["geometry", "bbox"]
    
    var localStorage: LocalStorage? = nil
    
    init(localStorage: LocalStorage){
        self.localStorage = localStorage
    }
    
    private func getEntityAttributeType(entity: NSEntityDescription, attrName: String)->NSAttributeType{
        let attributes = entity.attributesByName
        let attributeDescription = attributes[attrName]! as NSAttributeDescription
        return attributeDescription.attributeType
    }
    
    // MARK: - OUTPUT MAPPING (Core Data -> JSON)
    func getRequestJSON(item: Model, mapping: NSDictionary)->String{
        let result = getRequestRawJSON(item, mapping: mapping)
        return makeJSONString(result)
    }
    
    func getRequestRawJSON(item: Model, mapping: NSDictionary)->NSMutableDictionary{
        let result = iterateMappingToJSON(item, mapping: mapping)
        let type = item.getEventType()
        result.setValue(type, forKey: "type")
        return result
    }
    
    private func iterateMappingToJSON(item: NSManagedObject, mapping: NSDictionary)->NSMutableDictionary{
        let output = NSMutableDictionary()
        for (jsonKey, localKey) in mapping{
            let sub = localKey as? NSDictionary
            if(sub != nil){
                let subTree = iterateMappingToJSON(item, mapping: localKey as! NSDictionary)
                output.setValue(subTree, forKey: jsonKey as! String)
            }else{
                let localValue: AnyObject? = getConvertedLocalValue(item, key: localKey as! String)
                if(localValue != nil){
                    output.setValue(localValue, forKey: jsonKey as! String)
                }
            }
        }
        return output
    }
    
    private func getConvertedLocalValue(item: NSManagedObject, key: String)->AnyObject?{
        let range = key.rangeOfString(RELATIONAL_ENTITY_DELIMITER)
        if(range != nil){
            // related item value
            
            // split the key
            let relatedEntityName = key.substringToIndex(range!.startIndex)
            let attributeKey = key.substringFromIndex(range!.startIndex.successor())
            
            // get related object and value
            let relatedItem = item.valueForKey(relatedEntityName) as! NSManagedObject
            let value: AnyObject? = relatedItem.valueForKey(attributeKey)
            
            //convert
            return convertToJSON(relatedItem, key: attributeKey, object: value!)
            
        }else{
            // direct item value
            let value: AnyObject? = item.valueForKey(key)
            if(value != nil){
                return convertToJSON(item, key: key, object: value!)
            }
        }
        return nil
    }
    
    internal func convertToJSON(item: NSManagedObject, key: String, object: AnyObject)->AnyObject?{
        let attrType = getEntityAttributeType(item.entity, attrName: key)
        if(attrType == NSAttributeType.StringAttributeType){
            if(jsonStringAttributes.contains(key)){
                // convert string JSON to JSON array
                let jsonString = object as! String
                if let dataFromString = jsonString.dataUsingEncoding(NSUTF8StringEncoding, allowLossyConversion: false) {
                    let json = JSON(data: dataFromString).arrayObject
                    return json == nil ? [] : json!
                }
            }else{
                return object as! String
            }
        }else if(attrType == NSAttributeType.BooleanAttributeType){
            return object as! Bool
        }else if(attrType == NSAttributeType.DateAttributeType){
            return dateToJSON(object as! NSDate)
        }else if(attrType == NSAttributeType.BinaryDataAttributeType && arrayAttributes.contains(key)){
            // binary array
            return NSKeyedUnarchiver.unarchiveObjectWithData(object as! NSData)!
        }
        return object
    }
    
    // MARK: - INPUT MAPPING (JSON -> CoreData)
    
    func createLocal(json: JSON, entityName: String, mapping: NSDictionary)->NSManagedObject{
        let item = localStorage?.newEntity(entityName)
        iterateMappingToLocal(item!, mapping: mapping, json: json)
        localStorage?.create(item!)
        return item!
    }
    
    func createOrUpdateLocal(item: NSManagedObject?, json: JSON, entityName: String, mapping: NSDictionary)->NSManagedObject{
        var editedItem = item
        
        if(editedItem == nil){ // if nil then create a new one
            editedItem = localStorage?.newEntity(entityName)
            iterateMappingToLocal(editedItem!, mapping: mapping, json: json)
            localStorage?.create(editedItem!)
        }else if(json["hash"].stringValue > (editedItem!.valueForKey("hashString") as? String)){ // update existing item if hash has changed on the server
            iterateMappingToLocal(editedItem!, mapping: mapping, json: json)
            localStorage?.update(editedItem!)
            // delete the obsolete update events in the queue
            let queue = SyncQueue.getInstance(localStorage!.managedObjectContext)
            queue.deleteEventsForObject(editedItem!)
        }
        
        return editedItem!
    }
    
    func updateLocal(item: NSManagedObject, json: AnyObject, mapping: NSDictionary){
        iterateMappingToLocal(item, mapping: mapping, json: JSON(json))
        localStorage?.update(item)
    }
    
    private func iterateMappingToLocal(item: NSManagedObject, mapping: NSDictionary, json: JSON){
        for (jsonKey, localKey) in mapping{
            let sub = localKey as? NSDictionary
            if(sub != nil){
                //NSLog("substree %@", jsonKey as String)
                iterateMappingToLocal(item, mapping: localKey as! NSDictionary, json: json[jsonKey as! String])
            }else{
                //NSLog("item %@ = %@", jsonKey as String, localKey as String)
                setLocalValue(item, key: localKey as! String, value: json[jsonKey as! String])
            }
        }
    }
    
    private func setLocalValue(item: NSManagedObject, key: String, value: JSON){
        if(value != nil){
            let range = key.rangeOfString(RELATIONAL_ENTITY_DELIMITER)
            if(range != nil){
                // related item value
                
                // split the key
                let relatedEntityName = key.substringToIndex(range!.startIndex)
                let attributeKey = key.substringFromIndex(range!.startIndex.successor())
                
                // get the related object and converted value
                let relatedItem = item.valueForKey(relatedEntityName) as? NSManagedObject
                if(relatedItem != nil){
                    let convertedValue: (AnyObject) = convertFromJSON(relatedItem!, key: attributeKey, object: value)
                    // set
                    relatedItem!.setValue(convertedValue, forKey: attributeKey)
                }
            }else{
                //direct item value
                let convertedValue: (AnyObject) = convertFromJSON(item, key: key, object: value)
                return item.setValue(convertedValue, forKey: key)
            }
        }
    }

    internal func convertFromJSON(item: NSManagedObject, key: String, object: JSON)->AnyObject{
        let attrType = getEntityAttributeType(item.entity, attrName: key)
        if(attrType==NSAttributeType.DateAttributeType){
            return dateFromJSON(object.stringValue)
        }else if(attrType==NSAttributeType.StringAttributeType){
            if(jsonStringAttributes.contains(key)){
                // convert JSON array to JSON string
                let str = object.rawString(NSUTF8StringEncoding, options: NSJSONWritingOptions.PrettyPrinted)
                return str == nil ? "" : str!
            }else{
                return object.stringValue
            }
        }else if(attrType == NSAttributeType.BooleanAttributeType){
            return object.boolValue
        }else if(attrType == NSAttributeType.BinaryDataAttributeType && arrayAttributes.contains(key)){
            // binary array
            if(object.arrayObject == nil){
                return NSData()
            }else{
                let data = object.arrayObject! as [AnyObject]
                return NSKeyedArchiver.archivedDataWithRootObject(data)
            }
        }
        return object.object
    }

    // Relations mapping
    /** Creates relations to objects by name that is a part of URI. */
    func saveRelationsByURL(item: NSManagedObject, url: String, mapping: NSDictionary?){
        if(mapping != nil){
            let urlParts = (url as NSString).pathComponents
            
            for (index, entity) in mapping!{
                let storageName = (entity as! String) + "Storage"
                let storage = StorageFactory.createInstance(storageName, managedObjectContext: localStorage!.managedObjectContext)
                let relatedItem = storage!.getByKey(urlParts, index: index as! Int)
                item.setValue(relatedItem, forKey: (entity as! String).lowercaseString)
                localStorage!.saveChanges()
            }
        }
    }
    
    /** Creates relations to objects by name received as one of items in the received JSON. */
    func saveRelationsByAttr(item: NSManagedObject, receivedJSON: JSON, mapping: NSDictionary?){
        if(mapping != nil){
            for (index, entity) in mapping!{
                let storageName = (entity as! String) + "Storage"
                let storage = StorageFactory.createInstance(storageName, managedObjectContext: localStorage!.managedObjectContext)
                let relatedItem = storage!.getByName(receivedJSON[index as! String].stringValue)
                item.setValue(relatedItem, forKey: (entity as! String).lowercaseString)
                localStorage!.saveChanges()
            }
        }
    }
    
    
}
