//
//  FeatureDetailController.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 18/01/16.
//  Copyright © 2016 MENDELU. All rights reserved.
//

import Foundation
import UIKit
import SwiftyJSON
import GoogleMaps
import SwiftValidator
import CoreData
import UIImage_Resize

protocol FeatureDetailControllerDelegate{
    func didChangeFeature()
    func didCancelNewFeature()
}

class FeatureDetailController: BaseTableViewController, FeatureInfoControllerDelegate, FeatureImageCellDelegate, FeatureActionsCellDelegate, FeatureDeleteCellDelegate, DefaultFormValidatorDelegate, UIImagePickerControllerDelegate, UINavigationControllerDelegate, UIPopoverPresentationControllerDelegate, DownloadPictureControllerDelegate{
    
    let DEFAULT_PREVIEW_ZOOM: Float = 10
    
    var feature: Feature? = nil
    var layerInProject: LayerInProject? = nil
    var actualZoom: Float? = nil
    var tempGeometry: Geometry? = nil
    var tempPrecision: Double? = nil
    var tempPicture: UIImage? = nil
    var hasChangedPicture: Bool = false
    private var newFeatureCreated: NSDate = NSDate()
    
    var delegate: FeatureDetailControllerDelegate? = nil
    
    var managedObjectContext: NSManagedObjectContext? = nil
    var featureStorage: FeatureStorage? = nil
    var featureAttributeStorage: FeatureAttributeStorage? = nil
    var layerAttributeStorage: LayerAttributeStorage? = nil
    var featurePictureStorage: FeaturePictureStorage? = nil
    
    var createdText = ""
    var modifiedText = ""
    @IBOutlet weak var mapView: GMSMapView!
    
    var map: MapManager? = nil
    var layerAttributes: [LayerAttribute]? = nil
    var featureAttributes: [FeatureAttribute]? = nil
    var attributeCells: [FeatureAttributeCell] = []
    var imageCell: FeatureImageCell? = nil
    var validator: DefaultFormValidator? = nil
    var imagePicker: UIImagePickerController? = nil
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        map = MapManager(mapView: mapView)
        featurePictureStorage = FeaturePictureStorage(moc: managedObjectContext!)
        layerAttributeStorage = LayerAttributeStorage(moc: managedObjectContext!)
        
        loadValues()
        configureView()
        initValidation()
    }
    
    func configureView(){
        if(feature == nil){
            setTitleText(l("Add feature"))
        }else{
            setTitleText(l("Edit feature"))
        }
        if(isIPad()){
            mapView.hidden = true
            tableView.tableHeaderView = nil
        }else{
            loadFeatureToMap()
        }
    }
    
    private func initValidation(){
        validator = DefaultFormValidator()
    }
    
    func loadValues(){
        if(feature != nil){
            createdText = dateToJSON(feature!.created)
            modifiedText = dateToJSON(feature!.lastEdit)
            tempGeometry = Geometry(string: feature!.geometry, type: layerInProject!.layer!.kind)
            tempPrecision = feature!.precision
            loadPicture()
        }else{
            createdText = dateToJSON(newFeatureCreated)
            modifiedText = "-"
        }
    }
    
    func loadPicture(){
        tempPicture = feature?.picture?.image
    }
    
    func loadAttributes(){
        if(layerAttributes == nil){
            layerAttributes = layerAttributeStorage?.getAllForLayer(layerInProject!.layer!).fetchedObjects as? [LayerAttribute]
        }
        if(featureAttributes == nil){
            if(feature != nil){
                if feature!.attributes.count > 0{
                    featureAttributes = feature!.attributes.allObjects as? [FeatureAttribute]
                }
            }
        }
    }
    
    func loadFeatureToMap(){
        map!.removeAllMarkers()
        if(actualZoom == nil){
            actualZoom = DEFAULT_PREVIEW_ZOOM
        }
        if(tempGeometry != nil){
            map!.addTempFeature(tempGeometry!, lip: layerInProject!)
            let center = tempGeometry!.boundingBoxCenter
            map!.setPosition(center.latitude, lon: center.longitude, zoom: actualZoom!)
        }
    }
    
    func findFeatureAttribute(searched: LayerAttribute)->FeatureAttribute?{
        if(feature != nil){
            if(featureAttributes != nil){
                for attr in featureAttributes!{
                    if(attr.layerattribute == searched && attr.hidden != true){
                        return attr
                    }
                }
            }
        }
        return nil
    }
    
    func getFeatureType()->String{
        let activeLayer = layerInProject?.layer
        if(activeLayer!.kind == LayerTypes.point){
            return FeatureTypes.point
        }else if(activeLayer!.kind == LayerTypes.line){
            return FeatureTypes.line
        }else if(activeLayer!.kind == LayerTypes.polygon){
            return FeatureTypes.polygon
        }
        return ""
    }
    
    func saveFeature(){
        if(featureStorage != nil){
            
            if(feature != nil && feature?.managedObjectContext == nil){
                alert(l("Feature cannot be saved. Item has already been deleted!"))
                self.dismissViewControllerAnimated(true, completion: nil)
            }else{
            
                tempGeometry?.correctPolygonEnclosure()
                
                if(feature == nil){
                    // add new
                    feature = featureStorage!.newEntity() as? Feature
                    feature!.geometry = makeJSONString(tempGeometry?.data)
                    feature!.layer = layerInProject!.layer!
                    feature!.type = getFeatureType()
                    feature!.hasPicture = (tempPicture != nil)
                    feature!.precision = tempPrecision == nil ? 0 : tempPrecision!
                    feature!.created = newFeatureCreated
                    feature!.lastEdit = newFeatureCreated
                    if(layerInProject?.layer?.kind != LayerTypes.point){
                        feature!.bbox = makeJSONString(tempGeometry?.boundingBox)
                    }
                    featureStorage!.create(feature!)
                }else{
                    // edit existing
                    feature!.hasPicture =  (tempPicture != nil)
                    feature!.geometry = makeJSONString(tempGeometry?.data)
                    if(layerInProject?.layer?.kind != LayerTypes.point){
                        feature!.bbox = makeJSONString(tempGeometry?.boundingBox)
                    }
                    feature!.precision = tempPrecision == nil ? 0 : tempPrecision!
                    feature!.lastEdit = NSDate()
                    featureStorage!.update(feature!)
                }
                
                saveAttributes()
                
                if(hasChangedPicture){
                    savePicture()
                }
                
                NSLog("sync")
                featureStorage!.synchronise()
            }
            
        }else{
            alert("Saving error!")
        }
    }
    
    func saveAttributes(){
        if(layerAttributes != nil){
            for (index, layerAttribute) in layerAttributes!.enumerate(){
                if(layerAttribute.managedObjectContext != nil){
                    let featureAttribute = findFeatureAttribute(layerAttribute)
                    let cell = attributeCells[index]
                    let value = cell.valueField.text
                    
                    if(featureAttribute == nil){
                        NSLog("a")
                    }
                    if(value != nil){
                        NSLog("b")
                    }
                    if(value!.isEmpty == false){
                        NSLog("c")
                    }
                    if(featureAttribute == nil && value != nil && value!.isEmpty == false){
                        // add
                        let item = featureAttributeStorage!.newEntity() as! FeatureAttribute
                        item.valueString = value!
                        item.layerattribute = layerAttribute
                        item.feature = feature!
                        featureAttributeStorage?.create(item)
                        
                    }else if(featureAttribute != nil && (value == nil/* || value!.isEmpty*/)){
                        // delete
                        featureAttributeStorage?.delete(featureAttribute!)
                        
                    }else if(featureAttribute != nil && value != nil && value!.isEmpty == false && featureAttribute!.valueString != value!){
                        // update
                        featureAttribute!.valueString = value!
                        featureAttributeStorage?.update(featureAttribute!)
                    }
                }
            }
        }
    }
    
    func savePicture(){
        var featurePicture: FeaturePicture? = nil
        if(tempPicture != nil && feature?.picture == nil){
            // add
            featurePicture = featurePictureStorage?.newEntity() as? FeaturePicture
            featurePicture!.feature = feature
            featurePicture!.image = tempPicture
            featurePictureStorage!.create(featurePicture!)
            
        }else if(tempPicture != nil && feature?.picture != nil && feature?.picture?.managedObjectContext != nil){
            // update
            featurePicture = feature!.picture
            featurePicture!.image = tempPicture
            featurePictureStorage!.update(featurePicture!)
        }
    }
    
    func validationSuccessful() {
        saveFeature()
        delegate?.didChangeFeature()
        if(!isIPad()){
            dismissViewControllerAnimated(true, completion: nil)
        }
    }
    
    @IBAction func doneClick(sender: AnyObject) {
        validator!.validate(self)
    }
    
    @IBAction func closeClick(sender: AnyObject) {
        delegate?.didCancelNewFeature()
        if(!isIPad()){
            dismissViewControllerAnimated(true, completion: nil)
        }
    }
    
    @IBAction func addAttributeClick(sender: AnyObject) {
    }
    
    // MARK: - Feature Image
    func updatePictureControls(){
        let imagePath = NSIndexPath(forRow: 0, inSection: 1)
        tableView.reloadRowsAtIndexPaths([imagePath], withRowAnimation: UITableViewRowAnimation.Automatic)
    }
    
    @IBAction func addImageClick(sender: UIButton) {
        let optionMenu = UIAlertController(title: "Add image", message: "Choose an action", preferredStyle: UIAlertControllerStyle.ActionSheet)
        
        let action = UIAlertAction(
            title: l("Choose from gallery"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.addImageFromGallery(sender)
        })
        optionMenu.addAction(action)
        
        let action2 = UIAlertAction(
            title: l("Take a photo"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.addImageByCamera(sender)
        })
        optionMenu.addAction(action2)
        
        let cancelAction = UIAlertAction(title: l("Cancel"), style: .Cancel, handler: {
            (alert: UIAlertAction) -> Void in
        })
        optionMenu.addAction(cancelAction)
        
        // align the dialog to the right of the sender
        optionMenu.popoverPresentationController?.sourceView = sender
        optionMenu.popoverPresentationController?.sourceRect = sender.bounds
        
        // show the dialog
        presentViewController(optionMenu, animated: true, completion: nil)
    }
    
    func addImageFromGallery(sender: UIButton){
        imagePicker = UIImagePickerController()
        imagePicker!.allowsEditing = false
        imagePicker!.sourceType = .PhotoLibrary
        imagePicker!.delegate = self
        parentViewController!.presentViewController(imagePicker!, animated: true, completion: nil)
    }
    
    func addImageByCamera(sender: UIButton){
        if UIImagePickerController.availableCaptureModesForCameraDevice(.Rear) != nil {
            imagePicker = UIImagePickerController()
            imagePicker!.allowsEditing = false
            imagePicker!.sourceType = UIImagePickerControllerSourceType.Camera
            imagePicker!.cameraCaptureMode = .Photo
            imagePicker!.delegate = self
            presentViewController(imagePicker!, animated: true, completion: nil)
        } else {
            alert("This device does not have a camera!")
        }
    }
    
    func imagePickerController(picker: UIImagePickerController, didFinishPickingImage image: UIImage, editingInfo: [String : AnyObject]?) {
        imagePicker!.dismissViewControllerAnimated(true, completion: nil)
        imagePicker = nil
        
        // resize image
        let size = CGSize(width: config.featurePictureSize.width, height: config.featurePictureSize.height)
        let scaled = image.resizedImageToFitInSize(size, scaleIfSmaller: false)
        
        tempPicture = scaled
        hasChangedPicture = true
        updatePictureControls()
    }
    
    func imagePickerControllerDidCancel(picker: UIImagePickerController) {
        imagePicker!.dismissViewControllerAnimated(true, completion: nil)
        imagePicker = nil
    }
    
    func downloadImageClick(sender: UIButton) {
        let popoverContent = (self.storyboard?.instantiateViewControllerWithIdentifier("download-picture"))! as! DownloadPictureController
        popoverContent.modalPresentationStyle = UIModalPresentationStyle.Popover
        popoverContent.popoverPresentationController?.sourceView = sender
        popoverContent.popoverPresentationController?.sourceRect = sender.bounds
        popoverContent.popoverPresentationController!.delegate = popoverContent
        
        popoverContent.feature = feature
        popoverContent.featurePictureStorage = featurePictureStorage
        popoverContent.delegate = self
        
        parentViewController!.presentViewController(popoverContent, animated: true, completion: nil)
    }
    
    func didDownloadedPicture(picture: UIImage) {
        tempPicture = picture
        savePicture()
        updatePictureControls()
    }
    
    func imageClick(sender: UIButton) {
        let optionMenu = UIAlertController(title: "Feature Image", message: "Choose an action", preferredStyle: UIAlertControllerStyle.ActionSheet)
        
        let action = UIAlertAction(
            title: l("Show detail"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.showPictureDetail(sender)
        })
        optionMenu.addAction(action)
        
        let action2 = UIAlertAction(
            title: l("Share picture"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.sharePicture(sender)
        })
        optionMenu.addAction(action2)
        
        let action3 = UIAlertAction(
            title: l("Delete picture"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.deletePicture(sender)
        })
        optionMenu.addAction(action3)
        
        let action4 = UIAlertAction(
            title: l("Choose from gallery"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.addImageFromGallery(sender)
        })
        optionMenu.addAction(action4)
        
        let action5 = UIAlertAction(
            title: l("Take a new photo"),
            style: .Default,
            handler: {(alert: UIAlertAction) -> Void in
                self.addImageByCamera(sender)
        })
        optionMenu.addAction(action5)
        
        let cancelAction = UIAlertAction(title: l("Cancel"), style: .Cancel, handler: {
            (alert: UIAlertAction) -> Void in
        })
        optionMenu.addAction(cancelAction)
        
        // align the dialog to the right of the sender
        optionMenu.popoverPresentationController?.sourceView = sender
        optionMenu.popoverPresentationController?.sourceRect = sender.bounds
        
        // show the dialog
        presentViewController(optionMenu, animated: true, completion: nil)
    }
    
    func showPictureDetail(sender: UIButton){
        let popoverContent = (self.storyboard?.instantiateViewControllerWithIdentifier("feature-image-nav"))! as! UINavigationController
        let target = popoverContent.childViewControllers[0] as! PicturePreviewController
        target.picture = tempPicture
        
        parentViewController!.presentViewController(popoverContent, animated: true, completion: nil)
    }
    
    func sharePicture(sender: UIButton){
        if(tempPicture != nil){
            let objectsToShare = ["ahoj"]
            let activityVC = UIActivityViewController(activityItems: objectsToShare, applicationActivities: nil)
            
            // align the dialog to the right of the sender
            activityVC.popoverPresentationController?.sourceView = sender
            activityVC.popoverPresentationController?.sourceRect = sender.bounds
            
            self.presentViewController(activityVC, animated: true, completion: nil)
        }
    }
    
    func deletePicture(sender: UIButton){
        let optionMenu = UIAlertController(title: "Delete Feature Picture", message: "Do you really want to delete the picture pernamently?", preferredStyle: UIAlertControllerStyle.ActionSheet)
        
        let action = UIAlertAction(
            title: l("Delete"),
            style: .Destructive,
            handler: {(alert: UIAlertAction) -> Void in
                if(self.feature != nil && self.feature!.picture != nil){
                    self.featurePictureStorage?.delete(self.feature!.picture!)
                    //self.feature!.hasPicture = false
                    //self.featureStorage?.local.update(self.feature!) // do not sync, only local
                    self.featurePictureStorage?.synchronise()
                }
                self.tempPicture = nil
                self.updatePictureControls()
        })
        optionMenu.addAction(action)
        
        let cancelAction = UIAlertAction(title: l("Cancel"), style: .Cancel, handler: {
            (alert: UIAlertAction) -> Void in
        })
        optionMenu.addAction(cancelAction)
        
        // align the dialog to the right of the sender
        optionMenu.popoverPresentationController?.sourceView = sender
        optionMenu.popoverPresentationController?.sourceRect = sender.bounds
        
        // show the dialog
        presentViewController(optionMenu, animated: true, completion: nil)
    }
    
    // MARK: - Feature info
    
    func showFeatureInfo() {
        var name = "feature-info"
        if(layerInProject?.layer?.kind == LayerTypes.point.lowercaseString){
            name = "feature-info-point"
        }
        let target = self.storyboard?.instantiateViewControllerWithIdentifier(name) as! FeatureInfoController
        target.feature = feature
        target.tempGeometry = tempGeometry
        target.tempPrecision = tempPrecision
        target.newFeatureCreated = newFeatureCreated
        target.layerInProject = layerInProject
        target.delegate = self
        self.navigationController?.pushViewController(target, animated: true)
    }
    
    func editFeatureGeometry() {
        
    }
    
    func didChangeCoordinates(tempGeometry: Geometry?) {
        self.tempGeometry = tempGeometry
        loadFeatureToMap()
    }
    
    func deleteFeature(sender: UIButton) {
        let optionMenu = UIAlertController(title: "Delete feature", message: "Do you really want to delete this feature pernamently?", preferredStyle: UIAlertControllerStyle.ActionSheet)
        
        let action = UIAlertAction(
            title: l("Delete"),
            style: .Destructive,
            handler: {(alert: UIAlertAction) -> Void in
                if(self.feature != nil){
                    self.featureStorage?.delete(self.feature!)
                }
                self.delegate?.didChangeFeature()
                if(!isIPad()){
                    self.dismissView()
                }
        })
        optionMenu.addAction(action)
        
        let cancelAction = UIAlertAction(title: l("Cancel"), style: .Cancel, handler: {
            (alert: UIAlertAction) -> Void in
        })
        optionMenu.addAction(cancelAction)
        
        // align the dialog to the right of the row
        optionMenu.popoverPresentationController?.sourceView = sender
        optionMenu.popoverPresentationController?.sourceRect = sender.bounds
        
        // show the dialog
        presentViewController(optionMenu, animated: true, completion: nil)
    }
    
    // MARK: - Table view
    
    override func tableView(tableView: UITableView, titleForHeaderInSection section: Int) -> String? {
        switch section{
        case 0: return "Attributes"
        case 1: return "Image"
        case 2: return "Measurement"
        default: return nil
        }
    }
    
    override func numberOfSectionsInTableView(tableView: UITableView) -> Int {
        return 3
    }
    
    override func tableView(tableView: UITableView, numberOfRowsInSection section: Int) -> Int {
        switch section{
        case 0:
            loadAttributes()
            return layerAttributes!.count
        case 1: return 1
        case 2:
            if(feature == nil){
                return 3
            }else{
                return 4
            }
        default: return 1
        }
    }
    
    override func tableView(tableView: UITableView, heightForRowAtIndexPath indexPath: NSIndexPath) -> CGFloat {
        if(indexPath.section == 1 && (tempPicture != nil)){
            return 220
        }else{
            return super.tableView(tableView, heightForRowAtIndexPath: indexPath)
        }
    }
    
    override func tableView(tableView: UITableView, cellForRowAtIndexPath indexPath: NSIndexPath) -> UITableViewCell {
        switch indexPath.section{
            
        case 0:
            let cell = tableView.dequeueReusableCellWithIdentifier("AttributeCell", forIndexPath: indexPath) as! FeatureAttributeCell
            loadAttributes()
            let layerAttribute = layerAttributes![indexPath.row]
            cell.name.text = layerAttribute.title
            
            // number type
            if(layerAttribute.kind == AttributeKinds.number){
                cell.valueField.keyboardType = UIKeyboardType.DecimalPad
                validator?.registerField(cell.valueField, rules: [FloatRule(message: "beta")])
            }
            
            // value
            if let featureAttribute = findFeatureAttribute(layerAttribute){
                cell.valueField.text = featureAttribute.valueString
            }else if(layerAttribute.kind == AttributeKinds.number){
                cell.valueField.text = "0"
            }
            
            attributeCells.insert(cell, atIndex: indexPath.row)
            return cell
            
        case 1:
            let cell = tableView.dequeueReusableCellWithIdentifier("ImageCell", forIndexPath: indexPath) as! FeatureImageCell
            imageCell = cell
            cell.delegate = self
            // load picture and controls
            if(tempPicture != nil){
                imageCell!.mainImageView.hidden = false
                imageCell!.coverImageButton.hidden = false
                imageCell!.mainImageView.image = tempPicture
                imageCell!.downloadButton.enabled = false
                imageCell!.addImageButton.enabled = false
            }else{
                imageCell!.mainImageView.hidden = true
                imageCell!.coverImageButton.hidden = true
                imageCell!.downloadButton.enabled = true
                imageCell!.addImageButton.enabled = true
            }
            // hide download button for new feature or feature without image or feature with image to be deleted
            if(feature == nil || feature?.hasPicture == false){
                imageCell!.downloadButton.hidden = true
                imageCell!.downloadButton.enabled = false
            }
            return cell
            
        case 2:
            if(indexPath.row == 0){
                let cell = tableView.dequeueReusableCellWithIdentifier("CreatedCell", forIndexPath: indexPath) as! FeatureCreatedCell
                cell.createdText.text = createdText
                return cell
            }else if(indexPath.row == 1){
                let cell = tableView.dequeueReusableCellWithIdentifier("ModifiedCell", forIndexPath: indexPath) as! FeatureModifiedCell
                cell.modifiedText.text = modifiedText
                return cell
            }else if(indexPath.row == 2){
                let cell = tableView.dequeueReusableCellWithIdentifier("ActionsCell", forIndexPath: indexPath) as! FeatureActionsCell
                cell.delegate = self
                return cell
            }else if(indexPath.row == 3){
                let cell = tableView.dequeueReusableCellWithIdentifier("DeleteCell", forIndexPath: indexPath) as! FeatureDeleteCell
                cell.delegate = self
                return cell
            }
            
        default: return UITableViewCell()
        }
        return UITableViewCell()
    }
}
