//
//  LayerStorage.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 9/15/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import CoreData
import UIKit
import SwiftyJSON

class IncomingEventResolver: CentralStorage{
    
    let STATUS_CODE = 200
    let URL = "/sync/events"
    
    /// Array of changed storages. This array is used for notification of view controllers to refresh their data.
    var changedStorages: [String] = []
    
    /// Name used for NSNotifications of updated storages.
    static let NOTIFICATION_NAME = "receivedUpdatesFromServer"
    
    required init(moc: NSManagedObjectContext, delegate: NSFetchedResultsControllerDelegate? = nil) {
        super.init(moc: moc, delegate: delegate, entityName: "")
    }
    
    /**
     The request body contains an array of subjects for filtering events on the server.
     */
    override func prepareRequestData(event: Event)->SyncRequestData?{
        // subjects for filtering events - take attention of order!
        var data = [
            "/layer"
        ]
        
        // add currently openned project id
        let opennedProject = getAppDelegate().opennedProject
        if(opennedProject == nil){
            data.append("/project")
        }else{
            data.append("/project/"+opennedProject!.scheme+"/"+opennedProject!.name)
        }
        
        return SyncRequestData(event: event, url: URL, bodyString: makeJSONString(data))
    }
    
    /**
     Iterates received events.
     */
    override func requestSuccessful(request: SyncRequestData!, response: NSHTTPURLResponse!, data: AnyObject?){
        if(response.statusCode == STATUS_CODE && data != nil){
            let events = JSON(data!)
            for (_, event) in events{
                processEvent(event)
            }
            notifyViews()
        }
    }
    
    /**
     After all events were processed notify all relevant view controllers to refresh their data.
     */
    func notifyViews(){
        if(self.changedStorages.count > 0){
            let dc = NSNotificationCenter.defaultCenter()
            dc.postNotificationName(IncomingEventResolver.NOTIFICATION_NAME, object: self, userInfo: ["objects": self.changedStorages])
        }
    }
    
    /**
     Verifies and handles a change received from the server.
     
     - parameter event: Received event JSON object.
     */
    func processEvent(event: JSON){
        NSLog("incoming event = %@", makeJSONString(event.rawValue))
        
        // check
        if(event["method"] != nil && event["uri"] != nil){
            
            // get url parts
            let urlParts = NSURL(string: event["uri"].stringValue)!.pathComponents
            NSLog("url parts  = %@", urlParts!)
            
            // detect and init storage
            let storage = detectStorage(urlParts)
        
            // perform action
            if(storage != nil){
                handleAction(event, storage: storage!)
            }else{
                warning("Unrecognizable event type received!")
            }
            
        }else{
            warning("Wrong incoming event format!")
        }
    }
    
    
    /**
     Detects and inits a central storage subclass based on URL.
     
     - parameter urlParts: Array of String as URL path components.
     
     - returns: Initialized storage or nil, if the event can not be recognized (event is not implemented here).
     */
    func detectStorage(urlParts: [String]?)->CentralStorage?{
        var storage: CentralStorage? = nil
        
        // Project
        if(urlParts![1]=="project"){
            storage = ProjectStorage(moc: managedObjectContext, delegate: nil)
            
            // LayerInProject
            if(urlParts!.count==7 && urlParts![4]=="layer"){
                storage = LayerInProjectStorage(moc: managedObjectContext, delegate: nil)
            }
            
        // FeatureAttribute
        }else if((urlParts!.count == 7 || urlParts!.count == 8) && urlParts![4]=="feature" && urlParts![6]=="attribute"){
            storage = FeatureAttributeStorage(moc: managedObjectContext, delegate: nil)
            
        // FeaturePicture
        }else if(urlParts!.count == 7 && urlParts![4]=="feature" && urlParts![6]=="picture"){
            storage = FeaturePictureStorage(moc: managedObjectContext, delegate: nil)
            
        // Feature
        }else if((urlParts!.count == 5 || urlParts!.count == 6) && urlParts![4]=="feature"){
            storage = FeatureStorage(moc: managedObjectContext, delegate: nil)
            
        // LayerAttribute
        }else if((urlParts!.count == 5 || urlParts!.count == 6) && urlParts![1]=="layer" && urlParts![4]=="attribute"){
            storage = LayerAttributeStorage(moc: managedObjectContext, delegate: nil)
            
        // Layer
        }else if((urlParts!.count == 2 || urlParts!.count == 4) && urlParts![1]=="layer"){
            storage = LayerStorage(moc: managedObjectContext, delegate: nil)
        }
        
        return storage
    }
    
    /**
     Calls the appropriate incoming event handler in the given storage.
     
     - parameter event:   Received event JSON object.
     - parameter storage: The storage that should handle the event action.
     */
    func handleAction(event: JSON, storage: CentralStorage){
        switch(event["method"].stringValue){
            
        case HTTPMethod.POST:
            storage.handleIncomingPostEvent(event)
            
        case HTTPMethod.PUT:
            storage.handleIncomingPutEvent(event)
            
        case HTTPMethod.DELETE:
            storage.handleIncomingDeleteEvent(event)
            
        default:
            warning("Unrecognizable event method received!")
        }
        
        // success
        changedStorages.append(storage.entityName)
    }
}
