//
//  MapViewController.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 9/23/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import UIKit
import CoreData
import GoogleMaps
import SwiftyJSON

protocol MapControllerDelegate{
    func didChangeData()
}

class MapController: UIViewController, UISplitViewControllerDelegate, LayerInProjectControllerDelegate, UIPopoverPresentationControllerDelegate, CreateByWalkOptionsDelegate, MapManagerDelegate, UpdateNotificationDelegate, FeatureDetailControllerDelegate{
    
    var leftBarButton: UIBarButtonItem? = nil
    
    @IBOutlet weak var mapView: GMSMapView!
    @IBOutlet weak var mapScaleBar: MapScaleBarView!
    @IBOutlet weak var rightContainer: UIView!
    @IBOutlet weak var rightContainerWidth: NSLayoutConstraint!
    @IBOutlet weak var rightContainerZeroWidth: NSLayoutConstraint!
    
    @IBOutlet weak var targetCenterImage: UIImageView!
    @IBOutlet weak var targetHorizontalLine: UIView!
    @IBOutlet weak var targetVerticalLine: UIView!
    
    @IBOutlet weak var controlPanelView: UIView!
    @IBOutlet weak var manuallyButton: UIButton!
    @IBOutlet weak var automaticallyButton: UIButton!
    @IBOutlet weak var playButton: UIButton!
    @IBOutlet weak var controlMiddleLineView: UIView!
    
    
    var lockButton: UIBarButtonItem? = nil
    var cancelButton: UIBarButtonItem? = nil
    var doneButton: UIBarButtonItem? = nil
    var createByWalkButton: UIBarButtonItem? = nil
    var createManuallyButton: UIBarButtonItem? = nil
    var zoomToLayerButton: UIBarButtonItem? = nil
    var undoButton: UIBarButtonItem? = nil
    var flexibleSpace: UIBarButtonItem? = nil
    
    
    var managedObjectContext: NSManagedObjectContext? = nil
    var projectStorage: ProjectStorage? = nil
    
    var layerStorage: LayerStorage? = nil
    var layerInProjectStorage: LayerInProjectStorage? = nil
    var featureStorage: FeatureStorage? = nil
    var featureAttributeStorage: FeatureAttributeStorage? = nil
    
    var projectData: Project? = nil
    var activeLayerInProject: LayerInProject? = nil
    
    var delegate: MapControllerDelegate? = nil
    
    var createByWalkOptions: UIAlertController? = nil
    var map: MapManager? = nil
    var updateNotification: UpdateNotification? = nil
    var periodicMeasurement: NSTimer? = nil
    var periodicMeasurementPeriod: Int? = nil
    var tempGeometry = Geometry()
    var tempGeometryAccuracy: [Double?] = []
    
    var locked: Bool = true
    var measuring = false
    
    override func viewDidLoad() {
        featureAttributeStorage = FeatureAttributeStorage(moc: managedObjectContext!)
        
        initNavItems()
        initControlPanel()
        configureMap()
        loadFeatures()
        
        updateNotification = UpdateNotification(delegate: self)
        updateNotification!.registerForUpdates(["Feature", "LayerInProject", "Layer"])
    }
    
    override func viewWillAppear(animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.setToolbarHidden(false, animated: false)
        initActiveLayer()
        updateControlPanel()
    }
    
    override func viewWillDisappear(animated: Bool) {
        saveMapPosition()
        saveMapPreview()
        super.viewWillDisappear(animated)
    }
    
    private func initNavItems(){
        if(UIDevice.currentDevice().userInterfaceIdiom != .Pad){
            navigationItem.leftBarButtonItem = UIBarButtonItem(title: "Map layers", style: UIBarButtonItemStyle.Plain, target: self, action: "mapLayersPopOver:")
        }else{
            splitViewController?.preferredDisplayMode = UISplitViewControllerDisplayMode.PrimaryHidden
            splitViewController?.delegate = self
            leftBarButton = UIBarButtonItem(title: "Map layers", style: UIBarButtonItemStyle.Plain, target: self, action: "mapLayersOverlay:")
            navigationItem.leftBarButtonItem = leftBarButton!
        }
    }
    
    func initActiveLayer(){
        if(activeLayerInProject == nil){
            // Use temp storage to avoid conflicts with storage used in LayersInProjectController.
            let tempStorage = LayerInProjectStorage(moc: layerStorage!.managedObjectContext)
            let layersInProject = tempStorage.getAllForProject(projectData!)
            if layersInProject.fetchedObjects?.count>0{
                let firstLayerInProject = layersInProject.fetchedObjects![0] as! LayerInProject
                activeLayerInProject = firstLayerInProject
            }
        }
    }
    
    func initControlPanel(){
        initBarButtons()
        
        // make control panel rounded
        controlPanelView.layer.cornerRadius = 91.0;
        controlPanelView.layer.masksToBounds = true;
        
        // color button imagess according to button color
        let image = UIImage(named: Images.target)
        manuallyButton.setImage(image?.imageWithRenderingMode(UIImageRenderingMode.AlwaysTemplate), forState: .Normal)
        manuallyButton.tintColor = SwipeActionsColors.buttonBlue
        
        let image2 = UIImage(named: Images.satellite)
        automaticallyButton.setImage(image2?.imageWithRenderingMode(UIImageRenderingMode.AlwaysTemplate), forState: .Normal)
        automaticallyButton.tintColor = SwipeActionsColors.buttonBlue
    }
    
    func initBarButtons(){
        lockButton = UIBarButtonItem(image: UIImage(named: "Lock"), style: UIBarButtonItemStyle.Plain, target: self, action: "lockClick")
        cancelButton = UIBarButtonItem(title: "Cancel", style: UIBarButtonItemStyle.Plain, target: self, action: "cancelMeasurementClick")
        doneButton = UIBarButtonItem(title: "Done", style: UIBarButtonItemStyle.Done, target: self, action: "doneMeasurementClick")
        createByWalkButton = UIBarButtonItem(title: "Create by walk", style: UIBarButtonItemStyle.Plain, target: self, action: "createByWalkClick:")
        createManuallyButton = UIBarButtonItem(title: "Create manually", style: UIBarButtonItemStyle.Plain, target: self, action: "createManuallyClick")
        zoomToLayerButton = UIBarButtonItem(image: UIImage(named: "ZoomIn"), style: UIBarButtonItemStyle.Plain, target: self, action: "zoomToLayerClick")
        undoButton = UIBarButtonItem(image: UIImage(named: "Undo"), style: UIBarButtonItemStyle.Plain, target: self, action: "undoClick")
        
        // three spaces will be enough
        flexibleSpace = UIBarButtonItem(barButtonSystemItem: UIBarButtonSystemItem.FlexibleSpace, target: nil, action: nil)
    }
    
    private func configureMap(){
        map = MapManager(mapView: mapView)
        map!.delegate = self
        
        // return to the last postion
        if(projectData != nil && projectData?.lastLatitude != -1 && projectData?.lastLongitude != -1){
            map?.setPosition(projectData!.lastLatitude, lon: projectData!.lastLongitude, zoom: projectData!.lastZoomLevel)
        }
        
        // assign scale bar to MapManager
        map?.scaleBar = mapScaleBar
    }
    
    func loadFeatures(){
        map!.removeAllFeatures()
        //map!.reloadEditedFeature()
        featureStorage?.clearCache()
        let layers = projectData?.layers.allObjects as! [LayerInProject]
        for layerInProject in layers{
            if(layerInProject.visibility == true && layerInProject.hidden != true && layerInProject.layer != nil && layerInProject.layer!.hidden != true){
                //let features = layerInProject.layer!.features.allObjects as! [Feature] - nenacita zadne objekty po sync beginu, proto nahrazuji timto:
                let features = featureStorage?.getAllForLayer(layerInProject.layer!).fetchedObjects as! [Feature]
                for feature in features{
                    if(feature.hidden != true){
                        map!.addFeature(feature, lip: layerInProject)
                    }
                }
            }
        }
    }
    
    func didChangeData(){
        loadFeatures()
    }
    
    func didChangeFeature() {
        setRightPanelState(false)
        map!.removeEditingSymbol()
        map!.hideEditingMarkers()
        map!.saveTempFeature()
        loadFeatures()
        delegate?.didChangeData()
    }
    
    func didCancelNewFeature() {
        // cancel temp feature
        setRightPanelState(false)
        map!.removeEditingSymbol()
        map!.removeTempFeature()
    }
    
    func saveMapPosition(){
        let pos = map!.getPosition()
        projectData!.lastLatitude = pos.latitude
        projectData!.lastLongitude = pos.longitude
        projectData!.lastZoomLevel = pos.zoom!
        projectStorage?.update(projectData!)
    }
    
    /**
     Takes a screenshot of the map and saves it to the project as a project preview image.
     */
    func saveMapPreview(){
        //Create the UIImage
        UIGraphicsBeginImageContext(mapView.frame.size)
        mapView.layer.renderInContext(UIGraphicsGetCurrentContext()!)
        let image = UIGraphicsGetImageFromCurrentImageContext()
        UIGraphicsEndImageContext()
        
        // resize image
        let size = CGSize(width: ProjectCell.PREVIEW_IMAGE_WIDTH, height: ProjectCell.PREVIEW_IMAGE_HEIGHT)
        let scaled = image.resizedImageToSize(size)
        
        //Save it to the database
        projectData?.previewImage = scaled
        featureStorage?.local.saveChanges() // save changes in moc
    }
    
    func mapLayersPopOver(sender: AnyObject){
        let ctrl = self.storyboard?.instantiateViewControllerWithIdentifier("map-layers") as! UINavigationController

        let target2 = ctrl.viewControllers[0] as! LayersInProjectController
        if projectData != nil {
            target2.project = projectData
        }
        target2.managedObjectContext = managedObjectContext
        target2.layerStorage = layerStorage
        target2.storage = layerInProjectStorage
        target2.delegate = self
        target2.activeLayerInProject = activeLayerInProject
        
        ctrl.modalInPopover = true
        presentViewController(ctrl, animated: true,completion:  nil)
    }
    
    func mapLayersOverlay(sender: AnyObject){
        splitViewController?.preferredDisplayMode = UISplitViewControllerDisplayMode.AllVisible
        
    }
    
    func splitViewController(svc: UISplitViewController, willChangeToDisplayMode displayMode: UISplitViewControllerDisplayMode) {
        if(displayMode == UISplitViewControllerDisplayMode.AllVisible){
            navigationItem.leftBarButtonItem = nil
        }else{
            navigationItem.leftBarButtonItem = leftBarButton
        }
    }
    
    func didChangeActiveLayer(layerInProject: LayerInProject) {
        measuring = false
        periodicMeasurement?.invalidate()
        updateAutomaticMeasurementControls()
        activeLayerInProject = layerInProject
        locked = true
        updateControlPanel()
    }
    
    private func setTargetCrossVisibility(visible: Bool){
        targetCenterImage.hidden = !visible
        targetHorizontalLine.hidden = !visible
        targetVerticalLine.hidden = !visible
    }
    
    private func updateControlPanel(){
        if(locked){
            lockButton?.image = UIImage(named: "Lock")
            controlPanelView.hidden = true
            setTargetCrossVisibility(false)
            toolbarItems = [lockButton!, flexibleSpace!, zoomToLayerButton!]
            
            // stop and cancel actual measurement
            tempGeometry = Geometry()
            measuring = false
            //updateAutomaticMeasurementControls()
            //map!.removeEditingSymbol()
            map!.removeTempFeature()
            //map!.hideEditingMarkers()
            //setRightPanelState(false)
            
        }else if(activeLayerInProject != nil){
            lockButton?.image = UIImage(named: "Unlock")
            if(activeLayerInProject!.layer!.kind == LayerTypes.point.lowercaseString){
                setTargetCrossVisibility(true)
                controlPanelView.hidden = false
                controlPanelView.backgroundColor = UIColor.whiteColor()
                manuallyButton.hidden = false
                automaticallyButton.hidden = false
                controlMiddleLineView.hidden = false
                playButton.hidden = true
                toolbarItems = [lockButton!, flexibleSpace!, zoomToLayerButton!]
            }else{
                controlPanelView.hidden = true
                setTargetCrossVisibility(false)
                toolbarItems = [lockButton!, flexibleSpace!, createByWalkButton!, flexibleSpace!, createManuallyButton!, flexibleSpace!, zoomToLayerButton!]
            }
        }
    }
    
    func setRightPanelContent(ctrl: UIViewController){
        setRightPanelState(true)
        self.addChildViewController(ctrl)
        ctrl.view.frame = CGRectMake(0, 0, self.rightContainer.frame.size.width, self.rightContainer.frame.size.height);
        self.rightContainer.addSubview(ctrl.view)
        ctrl.didMoveToParentViewController(self)
    }
    
    // MARK: - ACTIONS
    
    @IBAction func closeButtonClick(sender: AnyObject) {
        dismissViewControllerAnimated(true, completion: nil)
    }
    
    func setRightPanelState(openned: Bool){
        rightContainerWidth.active = openned
        rightContainerZeroWidth.active = !openned
    }
    
    func didSelectFeature(feature: Feature, layerInProject: LayerInProject) {
        let navCtrl = self.storyboard?.instantiateViewControllerWithIdentifier("feature-detail-nav") as? UINavigationController
        let target = navCtrl!.childViewControllers[0] as? FeatureDetailController
        
        let position = map!.getPosition()
        target!.feature = feature
        target!.actualZoom = position.zoom
        target!.layerInProject = layerInProject
        target!.managedObjectContext = managedObjectContext
        target!.featureStorage = featureStorage
        target!.featureAttributeStorage = featureAttributeStorage
        target!.delegate = self
        
        if(isIPad()){
            setRightPanelContent(navCtrl!)
        }else{
            presentViewController(navCtrl!, animated: true, completion: nil)
        }
    }
    
    @IBAction func mapSwitchClick(sender: AnyObject) {
        let optionMenu = UIAlertController(title: nil, message: "Choose the map type", preferredStyle: UIAlertControllerStyle.ActionSheet)
        
        if(map?.mapTypes.count > 0){
            for type in map!.mapTypes{
                var title = l(type)
                let style = UIAlertActionStyle.Default
                
                // selected item
                if(title == map!.mapType){
                    title = "✓ " + title
                }
                
                let action = UIAlertAction(
                    title: title,
                    style: style,
                    handler: {(alert: UIAlertAction) -> Void in
                        self.map!.setType(type)
                })
                
                optionMenu.addAction(action)
            }
        }
        
        let cancelAction = UIAlertAction(title: l("Cancel"), style: .Cancel, handler: {
            (alert: UIAlertAction) -> Void in
        })
        optionMenu.addAction(cancelAction)
        
        optionMenu.popoverPresentationController?.sourceView = sender as? UIView
        optionMenu.popoverPresentationController?.sourceRect = sender.bounds
        
        presentViewController(optionMenu, animated: true, completion: nil)
    }
    
    func lockClick(){
        if(locked){
            if(activeLayerInProject != nil){
                locked = false
                updateControlPanel()
            }else{
                selectLayerAlert()
            }
        }else{
            locked = true
            updateControlPanel()
        }
    }
    
    func selectLayerAlert(){
        alert(l("Please select a layer first. If there is none in this map, create a new one first."))
    }
    
    // MARK: - MEASUREMENT ACTIONS
    
    func disabledLocationWarning(){
        alert(l("Please turn on location services!"))
    }
    
    @IBAction func automaticallyButtonClick(sender: AnyObject) {
        if let location = map!.getLocation(){
            // navigate to current location
            map!.setPosition(location.latitude, lon: location.longitude, zoom: map!.getPosition().zoom!)
            // add new point
            recordNewPosition(location)
        }else{
            disabledLocationWarning()
        }
    }
    
    @IBAction func manuallyButtonClick(sender: AnyObject) {
        let position = map!.getPosition()
        recordNewPosition(position)
    }
    
    func recordNewPosition(position: CameraPosition){
        if(activeLayerInProject?.layer!.kind == LayerTypes.point){
            let navCtrl = self.storyboard?.instantiateViewControllerWithIdentifier("feature-detail-nav") as! UINavigationController
            let target = navCtrl.childViewControllers[0] as! FeatureDetailController
            let geom = Geometry(json: [position.latitude, position.longitude], type: activeLayerInProject!.layer!.kind)
            target.tempGeometry = geom
            target.tempPrecision = position.horizontalAccuracy
            target.actualZoom = position.zoom
            target.layerInProject = activeLayerInProject
            target.managedObjectContext = managedObjectContext
            target.featureStorage = featureStorage
            target.featureAttributeStorage = featureAttributeStorage
            target.delegate = self
            
            if(isIPad()){
                setRightPanelContent(navCtrl)
                tempGeometry = geom
                map!.addTempFeature(tempGeometry, lip: activeLayerInProject!)
            }else{
                presentViewController(navCtrl, animated: true, completion: nil)
            }
        }else{
            tempGeometry.type = activeLayerInProject!.layer!.kind
            tempGeometry.append([position.latitude, position.longitude])
            tempGeometryAccuracy.append(position.horizontalAccuracy)
            map!.addTempFeature(tempGeometry, lip: activeLayerInProject!)
        }
    }

    func createByWalkClick(sender: UIBarButtonItem){
        let popoverContent = (self.storyboard?.instantiateViewControllerWithIdentifier("create-by-walk-options"))! as! CreateByWalkOptions
        popoverContent.modalPresentationStyle = UIModalPresentationStyle.Popover
        let popover = popoverContent.popoverPresentationController
        popover!.barButtonItem = sender
        popoverContent.popoverPresentationController!.delegate = self
        popoverContent.delegate = self

        self.presentViewController(popoverContent, animated: true, completion: nil)
    }
    
    func adaptivePresentationStyleForPresentationController(controller: UIPresentationController) -> UIModalPresentationStyle {
        return UIModalPresentationStyle.None
    }
    
    func startAutomaticMeasurement(period: Int){
        if let _ = map!.getLocation(){
            measuring = true
            periodicMeasurementPeriod = period
            updateAutomaticMeasurementControls()
            startTimer()
            automaticMeasurementTimerFired() // first point
        }else{
            disabledLocationWarning()
        }
    }
    
    func startTimer(){
        if(periodicMeasurementPeriod != nil){
            let interval = NSTimeInterval(periodicMeasurementPeriod!)
            periodicMeasurement = NSTimer.scheduledTimerWithTimeInterval(interval, target: self, selector: "automaticMeasurementTimerFired", userInfo: nil, repeats: true)
        }
    }

    func automaticMeasurementTimerFired() {
        if let location = map!.getLocation(){
            // add new point if different to the last one
            let last = tempGeometry.array.last
            if(last == nil || last![0] != location.latitude || last![1] != location.longitude){
                // navigate to current location
                map!.setPosition(location.latitude, lon: location.longitude, zoom: map!.getPosition().zoom!)
                // append new point to geometry
                recordNewPosition(location)
            }
        }else{
            //disabledLocationWarning()
        }
    }
    
    @IBAction func playButtonClick(sender: AnyObject) {
        if(measuring){
            measuring = false
            periodicMeasurement?.invalidate()
        }else{
            measuring = true
            startTimer()
        }
        updateAutomaticMeasurementControls()
    }
    
    func updateAutomaticMeasurementControls(){
        playButton.hidden = false
        manuallyButton.hidden = true
        automaticallyButton.hidden = true
        controlPanelView.hidden = false
        controlMiddleLineView.hidden = true
        toolbarItems = [cancelButton!, flexibleSpace!, doneButton!]
        if(measuring){
            playButton.setImage(UIImage(named: Images.pause), forState: UIControlState.Normal)
            controlPanelView.backgroundColor = SwipeActionsColors.delete
        }else{
            playButton.setImage(UIImage(named: Images.play), forState: UIControlState.Normal)
            controlPanelView.backgroundColor = SwipeActionsColors.duplicate
        }
    }
    
    func undoClick(){
        tempGeometry.popLast()
        map!.removeTempFeature()
        map!.addTempFeature(tempGeometry, lip: activeLayerInProject!)
    }
    
    func getAvgAccuracy()->Double{
        var sum = 0.0;
        for item in tempGeometryAccuracy{
            if(item != nil){
                sum += item!
            }
        }
        return sum / Double(tempGeometryAccuracy.count)
    }
    
    func doneMeasurementClick(){
        // check min count
        if(tempGeometry.array.count < 2 && activeLayerInProject!.layer!.kind == LayerTypes.line){
            alert(l("Linestring must have at least two points!"))
        }else if(tempGeometry.array.count < 3 && activeLayerInProject!.layer!.kind == LayerTypes.polygon){
            alert(l("Polygon must have at least three points!"))
        }else{
            // init form for adding new item
            let navCtrl = self.storyboard?.instantiateViewControllerWithIdentifier("feature-detail-nav") as! UINavigationController
            let target = navCtrl.childViewControllers[0] as! FeatureDetailController
            
            target.tempGeometry = tempGeometry
            target.tempPrecision = getAvgAccuracy()
            target.actualZoom = map!.getPosition().zoom
            target.layerInProject = activeLayerInProject
            target.managedObjectContext = managedObjectContext
            target.featureStorage = featureStorage
            target.featureAttributeStorage = featureAttributeStorage
            target.delegate = self
            
            // stop measuring
            measuring = false
            periodicMeasurement?.invalidate()
            tempGeometry = Geometry()
            updateControlPanel()
            
            // show form for adding new item
            if(isIPad()){
                setRightPanelContent(navCtrl)
            }else{
                presentViewController(navCtrl, animated: true, completion: nil)
            }
            
        }
    }
    
    func cancelMeasurementClick(){
        measuring = false
        tempGeometry = Geometry()
        map!.removeEditingSymbol()
        map!.removeTempFeature()
        periodicMeasurement?.invalidate()
        updateControlPanel()
    }
    
    func createManuallyClick(){
        controlPanelView.hidden = false
        manuallyButton.hidden = false
        automaticallyButton.hidden = false
        playButton.hidden = true
        controlMiddleLineView.hidden = false
        controlPanelView.backgroundColor = UIColor.whiteColor()
        setTargetCrossVisibility(true)
        toolbarItems = [cancelButton!, flexibleSpace!, undoButton!, flexibleSpace!, doneButton!]
    }
    
    func zoomToLayerClick(){
        if(activeLayerInProject == nil){
            selectLayerAlert()
        }else{
            map!.zoomToLayer(activeLayerInProject!)
        }
    }
    
    // MARK: - SEGUES
    override func prepareForSegue(segue: UIStoryboardSegue, sender: AnyObject?) {
        if(segue.identifier=="layers"){
            let navi = segue.destinationViewController as! UINavigationController
            let target = navi.viewControllers[0] as! LayersInProjectController
            target.managedObjectContext = managedObjectContext
            target.project = projectData
            target.storage = layerInProjectStorage
            target.layerStorage = layerStorage
        }
    }
}
