//
//  MapManager.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 9/23/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import GoogleMaps
import SwiftyJSON
import CoreLocation

protocol MapManagerDelegate{
    func didSelectFeature(feature: Feature, layerInProject: LayerInProject)
}

/**
 Class abstracting map services and controlling map objects.
*/
class MapManager: NSObject, CLLocationManagerDelegate, GMSMapViewDelegate{
    
    let locationManager = CLLocationManager()
    
    var mapView: GMSMapView
    var mapType: String
    var scaleBar: MapScaleBarView? = nil
    
    var delegate: MapManagerDelegate? = nil
    
    // all features array
    private var markers = [GMSMarker]()
    private var polylines = [GMSPolyline]()
    private var polygons = [GMSPolygon]()
    private var features = [String: MapObjectData]()
    
    // editing symbols
    var circleMarker: GMSMarker? = nil
    var editingMarkers = [GMSMarker]()
    var editedFeature: Feature? = nil
    
    // temp vars for new added features
    var tempFeatureMarker: GMSMarker? = nil
    var tempFeaturePolygon: GMSPolygon? = nil
    var tempFeaturePolyline: GMSPolyline? = nil
    
    var lastZoom: Float? = nil
    var minZIndex: Int32 = 0
    
    /// All available types of map bases
    var mapTypes = [
        "Topologic",
        "Satellite",
        "Hybrid",
        "Terrain",
        "None"
        ]
    
    init(mapView: GMSMapView){
        self.mapView = mapView
        mapType = mapTypes[0]
        
        super.init()
        
        setDefaults()
        locationManager.delegate = self
        locationManager.requestAlwaysAuthorization()
    }

    /**
        Sets up default map settings.

        - parameter bar: Consectetur adipisicing elit.
        - returns: Sed do eiusmod tempor.
     */
    func setDefaults(){
        mapView.myLocationEnabled = true
        mapView.settings.compassButton = true
        mapView.settings.myLocationButton = true
        mapView.delegate = self
    }
    
    /**
         Sets position of the map center.
         
         - parameter lat: Latitude
         - parameter lon: Longitude
         - parameter zoom: Zoom level
     */
    func setPosition(lat: Double, lon: Double, zoom: Float){
        let camera = GMSCameraPosition.cameraWithLatitude(lat, longitude: lon, zoom: zoom)
        mapView.camera=camera
    }
    
    /**
         Returns position of the map center.
         
         - returns: Position containing coordinates, zoom and accuracy.
     */
    func getPosition()->CameraPosition{
        let cam = mapView.camera
        return CameraPosition(latitude: cam.target.latitude, longitude: cam.target.longitude, zoom: cam.zoom, horizontalAccuracy: nil)
    }
    
    /**
     Returns the current location of the device based on GPS.
     
     - returns: Location as CameraPosition including horizontal accuracy. Zoom is always nil.
     */
    func getLocation()->CameraPosition?{
        if let location = mapView.myLocation{
            return CameraPosition(latitude: location.coordinate.latitude, longitude: location.coordinate.longitude, zoom: nil, horizontalAccuracy: location.horizontalAccuracy)
        }else{
            return nil
        }
    }
    
    func setType(type: String){
        let lType = type.lowercaseString
        switch lType{
        case "topologic": mapView.mapType = kGMSTypeNormal
        case "satellite": mapView.mapType = kGMSTypeSatellite
        case "hybrid": mapView.mapType = kGMSTypeHybrid
        case "terrain": mapView.mapType = kGMSTypeTerrain
        case "none": mapView.mapType = kGMSTypeNone
        default: mapView.mapType = kGMSTypeNormal
        }
        mapType = type
    }
    
    func addMarker(lat: Double, lon: Double) -> GMSMarker{
        let marker = GMSMarker()
        marker.position = CLLocationCoordinate2DMake(lat, lon)
        marker.zIndex = 2
        //marker.icon = GMSMarker.markerImageWithColor(UIColor.greenColor())
        markers.append(marker)
        return marker
    }
    
    func addImageMarker(lat: Double, lon: Double, image: UIImage)->GMSMarker{
        let marker = addMarker(lat, lon: lon)
        marker.icon = image
        return marker
    }
    
    func addEditingSymbol(marker: GMSMarker){
        hideEditingMarkers()
        removeEditingSymbol()
        let circle = UIImage(named: "yellow-circle1.png")
        circleMarker = GMSMarker()
        circleMarker!.position = CLLocationCoordinate2DMake(marker.position.latitude, marker.position.longitude)
        circleMarker!.map = mapView
        circleMarker!.icon = circle!
        circleMarker!.groundAnchor = CGPoint(x: 0.58, y: 0.79)
        circleMarker!.zIndex = minZIndex - 10
        markers.append(circleMarker!)
    }
    
    func reloadEditingSymbol(){
        if(editedFeature != nil && editedFeature?.managedObjectContext != nil){
            //removeEditingSymbol()
            //addEditingSymbol(editedFeature)
        }
    }
    
    func removeEditingSymbol(){
        circleMarker?.map = nil
        circleMarker = nil
    }
    
    func setEditingPolyline(polyline: GMSPolyline, editing: Bool){
        if(editing){
            //polyline.strokeWidth = 5
        }else{
            polyline.strokeWidth = 0
        }
    }
    
    func setEditingPolygon(polygon: GMSPolygon, editing: Bool){
        if(editing){
            //polygon.strokeWidth = 5
            //polygon.strokeColor = polygon.fillColor
        }else{
            polygon.strokeWidth = 0
        }
    }
    
    func showEditingMarkers(geometry: Geometry){
        hideEditingMarkers()
        removeEditingSymbol()
        for item in geometry.array{
            let marker = addMarker(item[0], lon: item[1])
            marker.map = mapView
            editingMarkers.append(marker)
        }
    }
    
    func hideEditingMarkers(){
        for (_, marker) in editingMarkers.enumerate(){
            marker.map = nil
            editingMarkers
                .removeAtIndex(0)
        }
    }
    
    func reloadEditingMarkers(){
        for (_, marker) in editingMarkers.enumerate(){
            marker.map = mapView
            markers.append(marker)
        }
    }
    
    func addPolyline(path: GMSPath) -> GMSPolyline{
        let polyline = GMSPolyline()
        polyline.path = path
        polyline.strokeWidth = 3
        polyline.zIndex = 2
        polyline.tappable = true
        polylines.append(polyline)
        return polyline
    }
    
    func addPolygon(path: GMSPath) -> GMSPolygon{
        let polygon = GMSPolygon()
        polygon.path = path
        polygon.zIndex = 2
        polygon.tappable = true
        polygons.append(polygon)
        return polygon
    }
    
    func addFeature(feature: Feature, lip: LayerInProject){
        let geometry = Geometry(string: feature.geometry, type: lip.layer!.kind)
        let featureData = MapObjectData(feature: feature, layerInProject: lip)
        let featureKey = lip.layer!.name + "," + feature.name
        let zIndex = Int32(-1 * lip.order.integerValue)
        minZIndex = min(minZIndex, zIndex)
        
        // point => marker
        if(lip.layer?.kind == LayerTypes.point.lowercaseString){
            if(geometry.isValidPoint()){
                let image = Symbols.getImage(lip.symbology, color: UIColor(hexString: lip.color!))
                let marker = addImageMarker(geometry.latitude, lon: geometry.longitude, image: image!)
                updateOverlayVisibility(marker, lip: lip)
                marker.zIndex = zIndex
                marker.userData = featureData
                featureData.object = marker
                if(editedFeature == feature){
                    addEditingSymbol(marker)
                }
            }else{
                warning("Unrecognizable geometry coordinates for point marker!")
            }
        
        // line => polyline
        }else if(lip.layer?.kind == LayerTypes.line.lowercaseString){
            let path = geometry.path
            let polyline = addPolyline(path)
            updateOverlayVisibility(polyline, lip: lip)
            polyline.zIndex = zIndex
            polyline.title = featureKey
            if(lip.color != nil){
                polyline.strokeColor = UIColor(hexString: lip.color!).colorWithAlphaComponent(CGFloat(lip.transparency.floatValue)/255)
            }
            featureData.object = polyline
        
        // polygon => polygon
        }else if(lip.layer?.kind == LayerTypes.polygon.lowercaseString){
            let path = geometry.path
            let polygon = addPolygon(path)
            updateOverlayVisibility(polygon, lip: lip)
            polygon.zIndex = zIndex
            polygon.title = featureKey
            if(lip.color != nil){
                polygon.fillColor = UIColor(hexString: lip.color!).colorWithAlphaComponent(CGFloat(lip.transparency.floatValue)/255)
            }
            featureData.object = polygon
        }
        
        // feature evidence array
        features[featureKey] = featureData
    }
    
    func addTempFeature(geometry: Geometry, lip: LayerInProject){
        removeTempFeature()
        
        let zIndex = Int32(-1 * lip.order.integerValue)
        minZIndex = min(minZIndex, zIndex)
        
        // point => marker
        if(lip.layer?.kind == LayerTypes.point){
            let image = Symbols.getImage(lip.symbology, color: UIColor(hexString: lip.color!))
            tempFeatureMarker = addImageMarker(geometry.latitude, lon: geometry.longitude, image: image!)
            tempFeatureMarker!.map = mapView
            tempFeatureMarker!.zIndex = zIndex
            addEditingSymbol(tempFeatureMarker!)
            
        // line => polyline
        }else if(lip.layer?.kind == LayerTypes.line){
            let path = geometry.path
            tempFeaturePolyline = addPolyline(path)
            tempFeaturePolyline!.map = mapView
            tempFeaturePolyline!.zIndex = zIndex
            if(lip.color != nil){
                tempFeaturePolyline!.strokeColor = UIColor(hexString: lip.color!).colorWithAlphaComponent(CGFloat(lip.transparency.floatValue)/255)
            }
            showEditingMarkers(geometry)
            setEditingPolyline(tempFeaturePolyline!, editing: true)
            
        // polygon => polygon
        }else if(lip.layer?.kind == LayerTypes.polygon){
            let path = geometry.path
            tempFeaturePolygon = addPolygon(path)
            tempFeaturePolygon!.map = mapView
            tempFeaturePolygon!.zIndex = zIndex
            if(lip.color != nil){
                tempFeaturePolygon!.fillColor = UIColor(hexString: lip.color!).colorWithAlphaComponent(CGFloat(lip.transparency.floatValue)/255)
            }
            showEditingMarkers(geometry)
            setEditingPolygon(tempFeaturePolygon!, editing: true)
            
        }
    }
    
    func reloadEditedFeature(){
        reloadTempFeature()
        reloadEditingSymbol()
        reloadEditingMarkers()
    }
    
    func reloadTempFeature(){
        if(tempFeatureMarker != nil){
            tempFeatureMarker?.map = mapView
            markers.append(tempFeatureMarker!)
        }
        if(tempFeaturePolyline != nil){
            tempFeaturePolyline?.map = mapView
            polylines.append(tempFeaturePolyline!)
        }
        if(tempFeaturePolygon != nil){
            tempFeaturePolygon?.map = mapView
            polygons.append(tempFeaturePolygon!)
        }
    }
    
    func saveTempFeature(){
        tempFeatureMarker = nil
        tempFeaturePolyline = nil
        tempFeaturePolygon = nil
        hideEditingMarkers()
    }
    
    func removeTempFeature(){
        tempFeatureMarker?.map = nil
        tempFeatureMarker = nil
        
        tempFeaturePolyline?.map = nil
        tempFeaturePolyline = nil
        
        tempFeaturePolygon?.map = nil
        tempFeaturePolygon = nil
        
        hideEditingMarkers()
    }
    
    func removeAllFeatures(){
        removeAllMarkers()
        removeAllPolylines()
        removeAllPolygons()
    }
    
    func removeAllMarkers(){
        for (_, marker) in markers.enumerate(){
            marker.map = nil
            markers.removeAtIndex(0)
        }
    }
    
    func removeAllPolylines(){
        for (_, polyline) in polylines.enumerate(){
            polyline.map = nil
            polylines.removeAtIndex(0)
        }
    }
    
    func removeAllPolygons(){
        for (_, polygon) in polygons.enumerate(){
            polygon.map = nil
            polygons.removeAtIndex(0)
        }
    }
    
    func updateFeaturesByThreshold(){
        for (_, item) in features{
            if(item.object != nil){
                updateOverlayVisibility(item.object!, lip: item.layerInProject)
            }
        }
    }
    
    func updateOverlayVisibility(overlay: GMSOverlay, lip: LayerInProject){
        let threshold = lip.zoomThreshold.floatValue
        let currentZoom = mapView.camera.zoom
        if(threshold > currentZoom || lip.visibility == false){
            overlay.map = nil
        }else{
            overlay.map = mapView
        }
    }
    
    func locationManager(manager: CLLocationManager, didChangeAuthorizationStatus status: CLAuthorizationStatus) {
        if(status != .AuthorizedAlways){
            //alert(l("Please allow location services!"))
        }
    }
    
    func mapView(mapView: GMSMapView!, didTapMarker marker: GMSMarker!) -> Bool {
        if let data = marker.userData as? MapObjectData{
            editedFeature = data.feature
            addEditingSymbol(marker)
            delegate?.didSelectFeature(data.feature, layerInProject: data.layerInProject)
            return true
        }else{
            return false
        }
    }
    
    func mapView(mapView: GMSMapView!, didTapOverlay overlay: GMSOverlay!) {
        if let data = features[overlay.title]{
            // linestring
            if(data.layerInProject.layer?.kind == LayerTypes.line){
                let polyline = overlay as! GMSPolyline
                let geometry = Geometry(string: data.feature.geometry, type: LayerTypes.line)
                setEditingPolyline(polyline, editing: true)
                showEditingMarkers(geometry)
                delegate?.didSelectFeature(data.feature, layerInProject: data.layerInProject)
            // polygon
            }else if(data.layerInProject.layer?.kind == LayerTypes.polygon){
                let polygon = overlay as! GMSPolygon
                let geometry = Geometry(string: data.feature.geometry, type: LayerTypes.polygon)
                setEditingPolygon(polygon, editing: true)
                showEditingMarkers(geometry)
                delegate?.didSelectFeature(data.feature, layerInProject: data.layerInProject)
            }
            
        }
    }
    
    func mapView(mapView: GMSMapView!, didChangeCameraPosition position: GMSCameraPosition!) {
        if(lastZoom == nil || lastZoom! != position.zoom){
            lastZoom = position.zoom
            updateFeaturesByThreshold()
        }
        scaleBar?.update(position.zoom, mapProjection: mapView.projection)
    }
    
    func zoomToLayer(lip: LayerInProject){
        
        // Calculate bounding box of all objects
        
        // default extreme values
        var minLat = 90.0
        var minLon = 180.0
        var maxLat = 0.0
        var maxLon = -180.0
        
        var hasFeature = false
        
        var firstFeatureLat: Double? = nil
        var firstFeatureLon: Double? = nil
        
        if(lip.layer?.kind == LayerTypes.point){
            // points
            for (_, item) in features{
                if(item.layerInProject == lip){
                    if let marker = item.object as? GMSMarker{
                        let geom = marker.position
                        // first item
                        if(firstFeatureLat == nil){
                            hasFeature = true
                            firstFeatureLat = geom.latitude
                            firstFeatureLon = geom.longitude
                        }
                        // min, max
                        if(geom.latitude < minLat){
                            minLat = geom.latitude
                        }
                        if(geom.longitude < minLon){
                            minLon = geom.longitude
                        }
                        if(geom.latitude > maxLat){
                            maxLat = geom.latitude
                        }
                        if(geom.longitude > maxLon){
                            maxLon = geom.longitude
                        }
                    }
                }
            }
        }else{
            // lines, polygons
            for (_, item) in features{
                if(item.layerInProject == lip && item.feature.bbox.isEmpty == false){
                    let geom = stringToJSON(item.feature.bbox)
                    if(geom.count == 4){
                        // first item
                        if(firstFeatureLat == nil){
                            hasFeature = true
                            firstFeatureLat = (geom[0].doubleValue + geom[2].doubleValue) / 2
                            firstFeatureLon = (geom[1].doubleValue + geom[3].doubleValue) / 2
                        }
                        // min, max
                        if(geom[0].doubleValue < minLat){
                            minLat = geom[0].doubleValue
                        }
                        if(geom[1].doubleValue < minLon){
                            minLon = geom[1].doubleValue
                        }
                        if(geom[2].doubleValue > maxLat){
                            maxLat = geom[2].doubleValue
                        }
                        if(geom[3].doubleValue > maxLon){
                            maxLon = geom[3].doubleValue
                        }
                    }
                }
            }
        }
        
        if(hasFeature){
            // define bounds
            let northEast = CLLocationCoordinate2D(latitude: maxLat, longitude: maxLon)
            let southWest = CLLocationCoordinate2D(latitude: minLat, longitude: minLon)
            let bounds = GMSCoordinateBounds(coordinate: northEast, coordinate: southWest)
            
            // Check min zoom threshold and update camera
            let camera = mapView.cameraForBounds(bounds, insets: UIEdgeInsets(top: 50, left: 50, bottom: 50, right: 50))
            let threshold = lip.zoomThreshold.floatValue
            if(camera.zoom < threshold && firstFeatureLat != nil && firstFeatureLon != nil){
                // use first feature position at min. zoom threshold
                mapView.camera = GMSCameraPosition.cameraWithLatitude(firstFeatureLat!, longitude: firstFeatureLon!, zoom: threshold)
            }else{
                // zoom to all objects of the current layer
                let update = GMSCameraUpdate.fitBounds(bounds, withPadding: 50)
                mapView.animateWithCameraUpdate(update)
            }
        }
    }
}
