//
//  PinLayerController.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 19/02/16.
//  Copyright © 2016 MENDELU. All rights reserved.
//

import Foundation
import UIKit
import CoreData

/// Controller for choosing and pinnning layers to a project.
class PinLayerController: LayersController{
    
    /// Layers are pinned to this project.
    var project: Project? = nil
    
    /// Used for creating new layer - project relationships.
    var layerInProjectStorage: LayerInProjectStorage? = nil
    
    /// Used for attaching order to new pinned layers.
    var lastOrder: Int = 0
    
    var cancelNavBarButton: UIBarButtonItem? = nil
    var pinNavBarButton: UIBarButtonItem? = nil
    
    override func viewDidLoad() {
        super.viewDidLoad()
        updateNotification!.registerForUpdates([]) // we do not want changes here
    }
    
    override func viewWillAppear(animated: Bool) {
        super.viewWillAppear(animated)
        navigationController?.setToolbarHidden(true, animated: false)
    }
    
    override func configureView() {
        // change nav bar buttons
        cancelNavBarButton = UIBarButtonItem(barButtonSystemItem: UIBarButtonSystemItem.Cancel, target: self, action: "cancelPinning:")
        pinNavBarButton = UIBarButtonItem(title: l("Pin"), style: UIBarButtonItemStyle.Done, target: self, action: "pinSelected:")
        self.navigationItem.leftBarButtonItem = cancelNavBarButton!
        self.navigationItem.rightBarButtonItem = pinNavBarButton!
        
        // show checkboxes for selecting layers
        tableView.setEditing(true, animated: false)
        
        // hide layer actions
        navigationController?.setToolbarHidden(true, animated: false)
    }
    
    /**
     Fetches all selected layers and pins them to the current project. The changes are synchronised at the end. Delegate didChangeData() is called and view is dismissed.
     */
    func pinSelected(sender: UIBarButtonItem){
        if(project != nil && layerInProjectStorage != nil){
            // iterate selected rows and fetch objects
            let paths = tableView.indexPathsForSelectedRows
            var objects = [Layer?]()
            for path in paths!{
                let object = storage?.get(path) as? Layer
                objects.append(object)
            }
            // iterate fetched objects
            for object in objects{
                // if layer exists and is not deleted, create a new relationship between layer and project (LayerInProject).
                if(object != nil && object!.managedObjectContext != nil){
                    createLayerInProject(object!)
                }
            }
            // sync
            layerInProjectStorage?.synchronise()
            delegate?.didChangeData()
        }
        dismissViewControllerAnimated(true, completion: nil)
    }
    
    /**
     Creates a new relation of the given layer and the current project with all default properties.
     
     - parameter layer: The layer to the relation.
     */
    private func createLayerInProject(layer: Layer){
        // init attributes
        let lip = layerInProjectStorage?.newEntity() as! LayerInProject
        lip.project = project
        lip.layer = layer
        lastOrder++
        lip.order = lastOrder
        lip.color = config.markers.colors[0]
        lip.symbology = config.markers.symbols[0]
        lip.transparency = 255
        lip.visibility = true
        lip.zoomThreshold = 7
        // create in storage
        layerInProjectStorage?.create(lip)
    }
    
    /**
     Hides the view controller.
     */
    func cancelPinning(sender: UIBarButtonItem){
        dismissViewControllerAnimated(true, completion: nil)
    }
    
    /**
     Filters only not pinned layers to the current project.
     */
    override func getAll() -> NSFetchedResultsController? {
        return (storage as! LayerStorage).getNotPinnedToProject(project!)
    }
}
