//
//  PokusController.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 16/02/16.
//  Copyright © 2016 MENDELU. All rights reserved.
//

import Foundation
import UIKit
import CoreData
import SwiftHTTP

protocol ProjectChangeDelegate{
    func didOpenProject(project: Project)
    func didCloseProject()
}

class ProjectsController: DatabaseCollectionViewController, WelcomePageControllerDelegate, ProjectCellDelegate{
    
    @IBOutlet weak var editButton: UIBarButtonItem!
    @IBOutlet weak var duplicateButton: UIBarButtonItem!
    @IBOutlet weak var deleteButton: UIBarButtonItem!
    
    var layerStorage: LayerStorage? = nil
    var layerInProjectStorage: LayerInProjectStorage? = nil
    var projectChangeDelegate: ProjectChangeDelegate? = nil
    var featureStorage: FeatureStorage? = nil
    
    override func viewDidLoad() {
        super.viewDidLoad()
        
        baseActions?.showWelcomeScreen(self)
        configureView()
        
        storage = ProjectStorage(moc: managedObjectContext!, delegate: nil)
        layerStorage = LayerStorage(moc: managedObjectContext!, delegate: nil)
        layerInProjectStorage = LayerInProjectStorage(moc: managedObjectContext!, delegate: nil)
        featureStorage = FeatureStorage(moc: managedObjectContext!)
        
        updateNotification!.registerForUpdates(["Project"])
    }
    
    override func viewWillAppear(animated: Bool) {
        super.viewWillAppear(animated)
        
        navigationController?.setToolbarHidden(true, animated: false)
        reloadData()
        
        // If the view is displayed, start delegating changes and reloading content.
        storage?.delegate = self
        
        // set openned project to nil - all projects are closed if we are here
        projectChangeDelegate?.didCloseProject()
    }
    
    override func viewWillDisappear(animated: Bool) {
        super.viewWillDisappear(animated)
        
        // Do not reload content if the view is not displayed (otherwise sync begin caused problems - no items were displayed, app restart needed).
        storage?.delegate = nil
    }
    
    override func didReceiveMemoryWarning() {
        super.didReceiveMemoryWarning()
        // Dispose of any resources that can be recreated.
    }
    
    func configureView(){
    }
    
    // MARK: - Local functions
    
    override func syncBeginSuccessful(){
        super.syncBeginSuccessful()
        storage?.clearCache()
        layerStorage?.clearCache()
        layerInProjectStorage?.clearCache()
        reloadData()
    }
    
    override func updateToolbarState(){
        let count = collectionView?.indexPathsForSelectedItems()?.count
        if(count>0 && editing){
            navigationController?.setToolbarHidden(false, animated: true)
            editButton.enabled=(count==1)
            duplicateButton.enabled=true
            deleteButton.enabled=true
        }else{
            navigationController?.setToolbarHidden(true, animated: true)
            editButton.enabled=false
            duplicateButton.enabled=false
            deleteButton.enabled=false
        }
    }
    
    /**
     Opens ProjectDetailController for editing the given project.
     
     - parameter indexPath: The indexPath of the item to be edited.
     */
    func editItem(object: Project?){
        if(object != nil && object?.managedObjectContext != nil){
            let navCtrl = self.storyboard?.instantiateViewControllerWithIdentifier("map-detail") as! UINavigationController
            let target = navCtrl.childViewControllers[0] as! ProjectDetailController
            target.storage = storage as? ProjectStorage
            target.project = object
            presentViewController(navCtrl, animated: true, completion: nil)
        }
    }
    
    /**
     Deletes the given project from the database. Includes a confirm dialog.
     
     - parameter object: Project to be deleted
     - parameter indexPath: The project cell
     */
    func deleteItem(object: Project?, indexPath: NSIndexPath?){
        if(indexPath != nil && object != nil && object?.managedObjectContext != nil){
            //self.collectionView.editing = false
            
            let optionMenu = UIAlertController(title: String(format: "Really delete map %@?", object!.title), message: l("All relevant content will be deleted too."), preferredStyle: .ActionSheet)
            
            let action = UIAlertAction(title: "Delete", style: .Destructive, handler: {
                (alert: UIAlertAction!) -> Void in
                self.storage!.delete(object!)
                self.storage!.synchronise()
            })
            
            let cancelAction = UIAlertAction(title: "Cancel", style: .Default, handler: {
                (alert: UIAlertAction!) -> Void in
            })
            
            optionMenu.addAction(action)
            optionMenu.addAction(cancelAction)
            
            let cell = collectionView(collectionView!, cellForItemAtIndexPath: indexPath!)
            optionMenu.popoverPresentationController?.sourceView = cell
            optionMenu.popoverPresentationController?.sourceRect = cell.bounds
            optionMenu.popoverPresentationController?.permittedArrowDirections = UIPopoverArrowDirection.Up
            
            presentViewController(optionMenu, animated: true, completion: nil)
        }
    }
    
    // MARK: - Selection related actions
    
    /**
     Click on the bottom edit button in selection related actions.
     */
    @IBAction func editClick(sender: AnyObject) {
        if let indexPaths = self.collectionView?.indexPathsForSelectedItems() {
            let object = storage?.get(indexPaths[0]) as? Project
            editItem(object)
        }
    }
    
    @IBAction func duplicateClick(sender: AnyObject) {
        alert("Not implemented yet!")
    }
    
    @IBAction func deleteClick(sender: AnyObject) {
        deleteSelected(sender as! UIBarButtonItem)
    }
    
    // MARK: - Other actions
    
    /**
     Click on the right bottom options button in the project cell.
     */
    func projectOptionsClick(cell: UICollectionViewCell, indexPath: NSIndexPath?) {
        let project = storage?.get(indexPath!) as! Project
        let optionMenu = UIAlertController(title: String(format: "Choose an action:"), message: project.title, preferredStyle: .ActionSheet)
        
        let action = UIAlertAction(title: "Edit", style: .Default, handler: {
            (alert: UIAlertAction!) -> Void in
            self.editItem(project)
        })
        let action2 = UIAlertAction(title: "Duplicate", style: .Default, handler: {
            (alert: UIAlertAction!) -> Void in
            
        })
        let action3 = UIAlertAction(title: "Delete", style: .Default, handler: {
            (alert: UIAlertAction!) -> Void in
            self.deleteItem(project, indexPath: indexPath)
        })
        
        let cancelAction = UIAlertAction(title: "Cancel", style: .Cancel, handler: {
            (alert: UIAlertAction!) -> Void in
        })
        
        optionMenu.addAction(action)
        optionMenu.addAction(action2)
        optionMenu.addAction(action3)
        optionMenu.addAction(cancelAction)
        
        optionMenu.popoverPresentationController?.sourceView = cell
        optionMenu.popoverPresentationController?.sourceRect = cell.bounds
        optionMenu.popoverPresentationController?.permittedArrowDirections = UIPopoverArrowDirection.Up
        
        presentViewController(optionMenu, animated: true, completion: nil)
    }
    
    /**
     Click on the left top Edit button for allowing multiple select.
     */
    @IBAction func switchEditingStateClick(sender: UIBarButtonItem) {
        deselectAll()
        
        if(editing == false){
            
            // start multiple editing
            
            editing = true
            collectionView!.allowsMultipleSelection = true
            
            sender.title = l("Done")
            sender.style = UIBarButtonItemStyle.Done
            
            // hide the tab bar and resize the main view to fill space after the hidden tab bar
            self.tabBarController?.tabBar.hidden = true
            var biggerFrame = tabBarController!.view.frame;
            biggerFrame.size.height += tabBarController!.tabBar.frame.size.height;
            tabBarController!.view.frame = biggerFrame ;

        }else{
            
            // end multiple editing
            
            editing = false
            collectionView!.allowsMultipleSelection = false
            
            sender.title = l("Edit")
            sender.style = UIBarButtonItemStyle.Plain
            
            // show the tab bar and resize the main view to fit with the displayed tab bar
            self.tabBarController?.tabBar.hidden = false
            var biggerFrame = tabBarController!.view.frame;
            biggerFrame.size.height -= tabBarController!.tabBar.frame.size.height;
            tabBarController!.view.frame = biggerFrame ;
            
        }
        
    }
    
    // MARK: - Segues
    override func shouldPerformSegueWithIdentifier(identifier: String, sender: AnyObject?) -> Bool {
        if(self.editing){
            
            return false
        }else{
            return super.shouldPerformSegueWithIdentifier(identifier, sender: sender)
        }
    }
    
    override func prepareForSegue(segue: UIStoryboardSegue, sender: AnyObject?) {
        if segue.identifier == "add-map" {
            let navi = segue.destinationViewController as! UINavigationController
            let target = navi.viewControllers[0] as! ProjectDetailController
            target.storage = storage as? ProjectStorage
        }
    }
    
    // MARK: - UICollectionView
    
    override func configureCell(cell: UICollectionViewCell, atIndexPath indexPath: NSIndexPath) {
        let mapCell = cell as! ProjectCell
        let object = storage!.get(indexPath) as! Project
        mapCell.titleText!.text = object.title
        
        if(object.descriptionString != ""){
            mapCell.descriptionText!.text = object.descriptionString
            mapCell.descriptionTextBg.hidden = false
        }else{
            mapCell.descriptionTextBg.hidden = true
        }
        
        mapCell.indexPath = indexPath
        mapCell.delegate = self
        
        if(object.previewImage != nil){
            mapCell.previewImage.image = object.previewImage!
        }else{
            mapCell.previewImage.image = UIImage(named: "default-map-preview.png")
        }
    }
    
    func deselectAll(){
        let all = collectionView?.indexPathsForSelectedItems()
        for indexPath in all!{
            collectionView?.deselectItemAtIndexPath(indexPath, animated: true)
        }
    }
    
    override func collectionView(collectionView: UICollectionView, didSelectItemAtIndexPath indexPath: NSIndexPath) {
        
        super.collectionView(collectionView, didSelectItemAtIndexPath: indexPath)
        
        if(collectionView.allowsMultipleSelection == false){
            
            // open selected project, if not editing mode enabled
            if(!editing){
                
                // clear cache
                layerInProjectStorage?.clearCache()
                
                // set the selected project as currently openned one
                if let indexPath = self.collectionView?.indexPathsForSelectedItems()![0] {
                    let project = storage?.get(indexPath) as! Project
                    projectChangeDelegate?.didOpenProject(project)
                }
                
                // different layouts - split for ipad, classic + popover for iphone
                if(isIPad()){
                    let ctrl = self.storyboard?.instantiateViewControllerWithIdentifier("map-split-view-controller") as! UISplitViewController
                    
                    // map
                    let navi2 = ctrl.viewControllers[1] as! UINavigationController
                    let target = navi2.viewControllers[0] as! MapController
                    if let indexPath = self.collectionView?.indexPathsForSelectedItems()![0] {
                        let object = storage?.get(indexPath) as! Project
                        target.projectData = object
                    }
                    target.managedObjectContext = managedObjectContext
                    target.projectStorage = storage as? ProjectStorage
                    target.layerStorage = layerStorage
                    target.layerInProjectStorage = layerInProjectStorage
                    target.featureStorage = featureStorage
                    
                    // layers in project
                    let target2Navi = ctrl.viewControllers[0] as! UINavigationController
                    let target2 = target2Navi.viewControllers[0] as! LayersInProjectController
                    if let indexPath = self.collectionView?.indexPathsForSelectedItems()![0] {
                        let object = storage?.get(indexPath) as! Project
                        target2.project = object
                    }
                    target2.managedObjectContext = managedObjectContext
                    target2.layerStorage = layerStorage
                    target2.storage = layerInProjectStorage
                    target2.featureStorage = featureStorage
                    target2.delegate = target
                    
                    target.delegate = target2
                    
                    presentViewController(ctrl, animated: true, completion: nil)
                }else{
                    //map
                    let ctrl = self.storyboard?.instantiateViewControllerWithIdentifier("map-view-controller") as! UINavigationController
                    let target = ctrl.viewControllers[0] as! MapController
                    if let indexPath = self.collectionView?.indexPathsForSelectedItems()![0] {
                        let object = storage?.get(indexPath) as! Project
                        target.projectData = object
                    }
                    target.managedObjectContext = managedObjectContext
                    target.projectStorage = storage as? ProjectStorage
                    target.layerStorage = layerStorage
                    target.layerInProjectStorage = layerInProjectStorage
                    target.featureStorage = featureStorage
                    
                    presentViewController(ctrl, animated: true, completion: nil)
                }
            }
        }else{
            let cell = collectionView.cellForItemAtIndexPath(indexPath) as! ProjectCell
            cell.previewImage.alpha = 0.6
        }
    }
    
    override func collectionView(collectionView: UICollectionView, didDeselectItemAtIndexPath indexPath: NSIndexPath) {
        super.collectionView(collectionView, didDeselectItemAtIndexPath: indexPath)
        
        if(collectionView.allowsMultipleSelection == true){
            let cell = collectionView.cellForItemAtIndexPath(indexPath) as! ProjectCell
            cell.previewImage.alpha = 1
        }
    }

    
    /* --- DEBUG TOOLS --- */
    
    override func collectionView(collectionView: UICollectionView, viewForSupplementaryElementOfKind kind: String, atIndexPath indexPath: NSIndexPath) -> UICollectionReusableView {
        if(Settings.debugMode && kind == UICollectionElementKindSectionFooter){
            let btn1: UIButton = UIButton()
            btn1.frame = CGRectMake(0, 10, 140, 20)
            btn1.setTitle("Clean the database", forState: UIControlState.Normal)
            btn1.backgroundColor = UIColor.clearColor()
            btn1.setTitleColor(UIColor.orangeColor(), forState: UIControlState.Normal)
            btn1.titleLabel?.font = UIFont.systemFontOfSize(13)
            btn1.addTarget(self, action: "cleanDatabase:", forControlEvents: UIControlEvents.TouchUpInside)
            
            let btnDeco: UIButton = UIButton()
            btnDeco.frame = CGRectMake(140, 10, 140, 20)
            btnDeco.setTitle("Run Sync Begin", forState: UIControlState.Normal)
            btnDeco.backgroundColor = UIColor.clearColor()
            btnDeco.setTitleColor(UIColor.orangeColor(), forState: UIControlState.Normal)
            btnDeco.titleLabel?.font = UIFont.systemFontOfSize(13)
            btnDeco.addTarget(self, action: "runSyncBegin:", forControlEvents: UIControlEvents.TouchUpInside)
            
            let btnChange: UIButton = UIButton()
            btnChange.frame = CGRectMake(0, 35, 140, 20)
            btnChange.setTitle("Reset Settings", forState: UIControlState.Normal)
            btnChange.backgroundColor = UIColor.clearColor()
            btnChange.setTitleColor(UIColor.orangeColor(), forState: UIControlState.Normal)
            btnChange.titleLabel?.font = UIFont.systemFontOfSize(13)
            btnChange.addTarget(self, action: "resetSettings:", forControlEvents: UIControlEvents.TouchUpInside)
            
            let footerView = collectionView.dequeueReusableSupplementaryViewOfKind(UICollectionElementKindSectionFooter, withReuseIdentifier: "FooterCell", forIndexPath: indexPath)
            //let footerView: UIView = UIView(frame: CGRectMake(20, 0, 280, 150))
            footerView.addSubview(btn1)
            footerView.addSubview(btnDeco)
            footerView.addSubview(btnChange)
            //footerView.frame = CGRectMake(footerView.frame.minX, footerView.frame.minY, 280, 50)
            //footerView.bounds = CGRectMake(0, 0, 280, 50)
            
            return footerView
        }
        return UICollectionReusableView()
    }
    
    func cleanDatabase(sender: UIButton){
        let sb = SyncBegin()
        sb.deleteAll()
    }
    
    func runSyncBegin(sender: UIButton){
        baseActions?.runSyncBegin()
    }
    
    func resetSettings(sender: UIButton){
        Settings.resetSettings(true)
        viewDidLoad()
    }
    
    /* --- END DEBUG TOOLS --- */
}
