//
//  ServerRequest.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 10/26/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import AFNetworking

class ServerRequest{
    
    var result: AnyObject?
    var operation: AFJSONRequestOperation? = nil
    var httpOperation: AFHTTPRequestOperation? = nil
    
    var method: String
    var url: String
    var body: NSData?
    let urlPrefix = Settings.URL
    var deviceID: Int? = nil
    
    /**
     Ctr for initializating a new server request.
     
     - parameter method: HTTP request method according to struct HTTPMethod.
     - parameter url:    Target URL as String.
     - parameter body:   HTTP body as NSData
     */
    init(method: String, url: String, body: NSData?){
        self.method = method
        self.url = url
        self.body = body
        
        if(Settings.deviceID != nil){
            deviceID = Settings.deviceID!
        }
    }
    
    /**
     Ctr for initializating a new server request using String for body.
     
     - parameter method: HTTP request method according to struct HTTPMethod.
     - parameter url:    Target URL as String.
     - parameter body:   HTTP body as String
     */
    convenience init(method: String, url: String, bodyString: String){
        let bodyData = ServerRequest.bodyStringToData(bodyString)
        self.init(method: method, url: url, body: bodyData)
    }
    
    /**
     Converts HTTP body string to NSData
     
     - parameter string: HTTP body as String
     
     - returns: HTTP body as NSData. If the conversion fails, returns nil.
     */
    static func bodyStringToData(body: String)->NSData?{
        return body.dataUsingEncoding(NSUTF8StringEncoding, allowLossyConversion: false)
    }
    
    /**
     Prepares request parameters like URL, device ID or Authorization field.
     
     - returns: Configured request.
     */
    private func prepareRequest()->NSMutableURLRequest{
        let validURL = sanitizeURL(urlPrefix! + "/" + url)
        let nsurl = NSURL(string: validURL)
        
        let request: NSMutableURLRequest = NSMutableURLRequest(URL: nsurl!)
        
        if(deviceID != nil){
            request.setValue(String(deviceID!), forHTTPHeaderField: "DEVICE_ID")
        }
        
        request.setValue(getAuthValue(), forHTTPHeaderField: "Authorization")
        request.HTTPMethod = method
        request.HTTPBody = body
        
        return request
    }
    
    /**
     Performs a server request. The response body is converted to JSON object.
     
     - parameter success: Success callback function. The last parameter is response body.
     - parameter failure: Failure callback function. The last parameter is response body.
     */
    func run(success: ((NSURLRequest!, NSHTTPURLResponse!, AnyObject!) -> Void)?, failure: ((NSURLRequest!, NSHTTPURLResponse!, NSError!, AnyObject!) -> Void)?){
        
        let request = prepareRequest()
        
        operation = AFJSONRequestOperation(
            request: request,
            success: success,
            failure: failure
        )
        
        //operation.waitUntilFinished()
        operation!.start()
    }
    
    /**
     Performs a request without converting response to JSON.
     
     - parameter success: Success callback function. The last parameter is response body.
     - parameter failure: Failure callback function. The last parameter is response body.
     */
    func runWithoutJSON(success: ((NSURLRequest!, NSHTTPURLResponse!, AnyObject!) -> Void)?, failure: ((NSURLRequest!, NSHTTPURLResponse!, NSError!, AnyObject!) -> Void)?){
        
        let request = prepareRequest()
        
        httpOperation = AFHTTPRequestOperation(request: request)
        httpOperation!.setCompletionBlockWithSuccess({ (operation:AFHTTPRequestOperation!, data:AnyObject!) -> Void in
            success!(request, operation.response, data)
            return Void()
            }, failure: { (operation:AFHTTPRequestOperation!, error:NSError!) -> Void in
                failure!(request, operation.response, error, nil)
                return Void()
        })
        
        httpOperation!.start()
    }
    
    /**
     Tries to stop the current request operation.
     */
    func stop(){
        operation?.cancel()
        httpOperation?.cancel()
    }
    
    /**
     Prepares data for login at server side using basic HTTP authorization.
     
     - returns: Encoded Authorization field for HTTP request header.
     */
    private func getAuthValue() -> String{
        let username = Settings.username
        let password = Settings.password
        let authStr: String = username!+":"+password!
        let authData: NSData = authStr.dataUsingEncoding(NSUTF8StringEncoding)!
        let authValue: String = "Basic \(authData.base64EncodedStringWithOptions([]))"
        return authValue
    }
}