//
//  SyncBegin.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 9/8/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import AFNetworking
import CoreData
import SwiftyJSON

class SyncBegin: NSObject {
    
    let URL = "/sync/begin"
    let RESPONSE_CODE = 200
    
    var moc: NSManagedObjectContext
    var model: NSManagedObjectModel
    var syncQueue: SyncQueue
    var successCallback: ((Void)->Void)? = nil
    var badLoginCallback: ((Void)->Void)? = nil
    
    /// If true, downloaded data is merged with the current database instead of replacing it.
    var useMerge: Bool
    
    private var data: AnyObject? = nil
    private var syncBeginTimestamp: Int64? = nil

    required init(useMerge: Bool = false){
        let a : AppDelegate = UIApplication.sharedApplication().delegate as! AppDelegate
        moc = a.managedObjectContext!
        model = a.managedObjectModel
        syncQueue = SyncQueue.getInstance(moc)
        self.useMerge = useMerge
        super.init()
    }
    
    func run(){
        // stop all ongoing sync events
        syncQueue.pause()
        // download JSON data
        downloadData()
    }
    
    func downloadData(){
        let request = ServerRequest(method: HTTPMethod.GET, url: URL, bodyString: "")
        request.deviceID = nil
        request.run({ (request: NSURLRequest!, response: NSHTTPURLResponse!, data: AnyObject!) -> Void in
            if(response.statusCode == self.RESPONSE_CODE){
                self.processData(data)
            }
            }, failure: {(request: NSURLRequest!, response: NSHTTPURLResponse!, error: NSError!, data: AnyObject!) -> Void in
                if(response == nil || response.statusCode == self.syncQueue.BAD_LOGIN_CODE){
                    self.badLoginCallback!()
                }
                return Void()
        })
    }
    
    func processData(data: AnyObject!){
        // stop all ongoing sync events
        syncQueue.pause()
        
        // delete old data and save new or do a merge
        if(data != nil){
            
            if(useMerge){
                NSLog("merge mode")
                let timestamp: Double = NSDate().timeIntervalSince1970
                syncBeginTimestamp = Int64(timestamp*1000.0);
            }else{
                NSLog("replacement mode")
                deleteAll()
            }
            
            let json = JSON(data)
            processDeviceID(json)
            processUserData(json)
            processLayers(json)
            processProjects(json)
            
            if(useMerge){
                deleteNotUpdatedObjects()
            }
        }
        
        // resume the common difference synchronisation
        syncQueue.resume()
        
        // send signal to the view controller
        self.successCallback?()
    }
    
    func deleteAll(){
        deleteAllEntities("Event", moc: moc)
        deleteAllEntities("Layer", moc: moc)
        deleteAllEntities("Project", moc: moc)
        deleteAllEntities("LayerInProject", moc: moc)
        deleteAllEntities("FeatureAttribute", moc: moc)
        deleteAllEntities("LayerAttribute", moc: moc)
        deleteAllEntities("Feature", moc: moc)
        deleteAllEntities("FeaturePicture", moc: moc)
        
    }
    
    /**
    Deletes all objects that were not modified in this sync begin run. It means that these objects has been already deleted at the server. Used for merge process.
    */
    func deleteNotUpdatedObjects(){
        deleteNotUpdatedObjects("Layer", moc: moc)
        deleteNotUpdatedObjects("Project", moc: moc)
        deleteNotUpdatedObjects("LayerInProject", moc: moc)
        deleteNotUpdatedObjects("FeatureAttribute", moc: moc)
        deleteNotUpdatedObjects("LayerAttribute", moc: moc)
        deleteNotUpdatedObjects("Feature", moc: moc)
        deleteNotUpdatedObjects("FeaturePicture", moc: moc)
        
    }
    
    func deleteNotUpdatedObjects(entityName: String, moc: NSManagedObjectContext){
        let fetchRequest = NSFetchRequest()
        fetchRequest.entity = NSEntityDescription.entityForName(entityName, inManagedObjectContext: moc)
        fetchRequest.includesPropertyValues = false
        fetchRequest.predicate = NSPredicate(format: "(mergeFlag!=%@) AND (hashString!='')", String(syncBeginTimestamp!))
        
        let error:NSError?
        do{
            let results = try moc.executeFetchRequest(fetchRequest) as? [NSManagedObject]
            for result in results! {
                moc.deleteObject(result)
            }
            
            do {
                try moc.save()
                // do something after save
                
            } catch let error1 as NSError {
                error = error1
                if let error = error {
                    print(error.userInfo)
                }
            }
            
        } catch {
            print("error: \(error)")
        }
    }
    
    /**
        Saves new generated device ID received from the server.
     
        - parameter data: Received complete JSON.
     */
    func processDeviceID(data: JSON){
        Settings.deviceID = data["device_id"].intValue
    }
    
    func processUserData(data: JSON){
        let user = data["user"]
        if(user != nil){
            Settings.name = user["name"].string
            Settings.email = user["email"].string
            Settings.phone = user["phone"].string
        }
    }
    
    func processLayers(data: JSON){
        let layers = data["layers"] as JSON
        let storage = LayerStorage(moc: moc, delegate: nil)
        let mapping = DataMapping(localStorage: storage.local)
        if(layers != nil){
            for (_, layerData) in layers{
                let layer: Layer?
                if(useMerge){
                    let item = storage.getByName(layerData["name"].stringValue)
                    layer = mapping.createOrUpdateLocal(item, json: layerData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? Layer
                    layer!.mergeFlag = syncBeginTimestamp!
                }else{
                    layer = mapping.createLocal(layerData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? Layer
                }
                processLayerAttributes(layer!, attributes: layerData["attributes"])
                processFeatures(layer!, features: layerData["features"])
            }
        }
    }
    
    func processLayerAttributes(layer: Layer, attributes: JSON){
        let storage = LayerAttributeStorage(moc: moc, delegate: nil)
        let mapping = DataMapping(localStorage: storage.local)
        if(attributes != nil){
            for (_, attributeData) in attributes{
                var attribute: LayerAttribute?
                if(useMerge){
                    let item = storage.getByLayerAndName(layer.name, attrName: attributeData["name"].stringValue)
                    attribute = mapping.createOrUpdateLocal(item, json: attributeData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? LayerAttribute
                    attribute!.mergeFlag = syncBeginTimestamp!
                }else{
                    attribute = mapping.createLocal(attributeData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? LayerAttribute
                }
                attribute!.layer = layer
            }
        }
    }
    
    func processFeatures(layer: Layer, features: JSON){
        let storage = FeatureStorage(moc: moc, delegate: nil)
        let mapping = DataMapping(localStorage: storage.local)
        if(features != nil){
            for (_, featureData) in features{
                let feature: Feature?
                if(useMerge){
                    let item = storage.getByLayerAndFeature(layer.name, featureName: featureData["name"].stringValue)
                    feature = mapping.createOrUpdateLocal(item, json: featureData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? Feature
                    feature!.mergeFlag = syncBeginTimestamp!
                }else{
                    feature = mapping.createLocal(featureData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? Feature
                }
                feature!.layer = layer
                processFeatureAttributes(feature!, attributeValues: featureData["attributes"])
            }
        }
    }
    
    func processFeatureAttributes(feature: Feature, attributeValues: JSON){
        let storage = FeatureAttributeStorage(moc: moc, delegate: nil)
        let attributeStorage = LayerAttributeStorage(moc: moc, delegate: nil)
        let mapping = DataMapping(localStorage: storage.local)
        if(attributeValues != nil){
            for (_, attributeValueData) in attributeValues{
                if(attributeValueData["value"].stringValue != ""){
                    let attributeValue: FeatureAttribute?
                    if(useMerge){
                        let item = storage.getByLayerAndFeatureAndAttribute(feature.layer.name, featureName: feature.name, attrName: attributeValueData["name"].stringValue)
                        attributeValue = mapping.createOrUpdateLocal(item, json: attributeValueData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? FeatureAttribute
                        attributeValue!.mergeFlag = syncBeginTimestamp!
                    }else{
                        attributeValue = mapping.createLocal(attributeValueData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? FeatureAttribute
                    }
                    attributeValue!.feature = feature
                    attributeStorage.clearCache()
                    let attribute = attributeStorage.getByName(attributeValueData["name"].string) as! LayerAttribute?
                    if(attribute != nil){
                        attributeValue!.layerattribute = attribute!
                    }else{
                        
                    }
                }
            }
        }
    }
    
    func processProjects(data: JSON){
        let projects = data["projects"]
        let storage = ProjectStorage(moc: moc, delegate: nil)
        let mapping = DataMapping(localStorage: storage.local)
        if(projects != nil){
            for (_, projectData) in projects{
                var project: Project?
                if(useMerge){
                    let item = storage.getByName(projectData["name"].stringValue)
                    project = mapping.createOrUpdateLocal(item, json: projectData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? Project
                    project!.mergeFlag = syncBeginTimestamp!
                }else{
                    project = mapping.createLocal(projectData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? Project
                }
                processLayersInProject(project!, layers: projectData["layers"])
            }
        }
    }
    
    func processLayersInProject(project: Project, layers: JSON){
        let storage = LayerInProjectStorage(moc: moc, delegate: nil)
        let layerStorage = LayerStorage(moc: moc, delegate: nil)
        let mapping = DataMapping(localStorage: storage.local)
        if(layers != nil){
            for (_, layerInProjectData) in layers{
                var layerInProject: LayerInProject?
                if(useMerge){
                    let item = storage.getByProjectAndLayer(project.name, layerName: layerInProjectData["layer"]["name"].stringValue)
                    layerInProject = mapping.createOrUpdateLocal(item, json: layerInProjectData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? LayerInProject
                    layerInProject!.mergeFlag = syncBeginTimestamp!
                }else{
                    layerInProject = mapping.createLocal(layerInProjectData, entityName: storage.ENTITY_NAME, mapping: storage.getMapping()!) as? LayerInProject
                }
                layerStorage.clearCache()
                let layer = layerStorage.getByName(layerInProjectData["layer"]["name"].string) as! Layer?
                if(layer != nil){
                    layerInProject!.project = project
                    layerInProject!.layer = layer!
                }
            }
        }
    }
    
}
