//
//  ServerStorage.swift
//  Gisella
//
//  Created by Ivo Pisarovic on 9/15/15.
//  Copyright (c) 2015 MENDELU. All rights reserved.
//

import Foundation
import CoreData
import AFNetworking
import SwiftyJSON

struct SyncQueueInstance{
    static let sharedInstance = SyncQueue()
}

class SyncQueue: SyncBeginControllerDelegate{
    
    let ACCEPTABLE_RESPONSE_CODES = [200, 201, 202, 204, 404, 409]
    let BAD_LOGIN_CODE = 401
    let DEVICE_ID_EXPIRED_CODE = 460
    
    var managedObjectContext: NSManagedObjectContext? = nil
    var entity: NSEntityDescription? = nil
    var syncing = false
    var stop = false
    
    /// If true, error info dialog will not be shown again. Prevents multiple dialogs at once.
    var hideErrorDialog = false
    
    /// If true, bad login info dialog will not be shown again. Prevents multiple dialogs at once.
    var hideBadLoginDialog = false
    
    // If true, updates are downloaded from the server after all request in the queue are successfully processed. Then it is automatically changed to 'false' before the end of this synchronisation process.
    var checkServerUpdates = false
    
    /* singleton */
    class func getInstance(managedObjectContext: NSManagedObjectContext? = nil)->SyncQueue{
        if(managedObjectContext != nil){
            SyncQueueInstance.sharedInstance.managedObjectContext = managedObjectContext
            SyncQueueInstance.sharedInstance.entity = NSEntityDescription.entityForName(Event.getEntityName(), inManagedObjectContext: managedObjectContext!)!
        }
        return SyncQueueInstance.sharedInstance
    }

    /* Blocks direct initialization, this is a singleton, use SyncQueue.getInstance() instead. */
    private init(){
        
    } 
    
    func createEvent(method: String, handlerClass: String?, itemId: NSURL?)->Event{
        let event = Event(entity: entity!, insertIntoManagedObjectContext: managedObjectContext)
        event.time=NSDate()
        event.method=method
        
        if(handlerClass != nil){
            event.handlerClass = handlerClass!
        }
        
        if(itemId != nil){
            event.itemPointer = itemId
        }
        
        var error: NSError? = nil
        do {
            try managedObjectContext!.save()
        } catch let error1 as NSError {
            error = error1
            // Replace this implementation with code to handle the error appropriately.
            // abort() causes the application to generate a crash log and terminate. You should not use this function in a shipping application, although it may be useful during development.
            //println("Unresolved error \(error), \(error.userInfo)")
            abort()
        }
        
        return event
    }
    
    func deleteEvent(event: Event){
        managedObjectContext!.deleteObject(event)
        var error: NSError? = nil
        do {
            try managedObjectContext!.save()
        } catch let error1 as NSError {
            error = error1
            print("Unresolved error \(error), \(error?.description)", terminator: "")
            abort()
        }
    }
    
    /**
    Deletes all events that are targeting to the given object.
    */
    func deleteEventsForObject(object: NSManagedObject){
        let fetchRequest = NSFetchRequest()
        fetchRequest.entity = NSEntityDescription.entityForName(Event.getEntityName(), inManagedObjectContext: managedObjectContext!)
        do{
            let results = try managedObjectContext!.executeFetchRequest(fetchRequest) as? [NSManagedObject]
            if(results!.count>0){
                for result in results!{
                    let event = result as! Event
                    var item: NSManagedObject? = nil
                    if(event.itemPointer != nil){
                        item = getManagedObjectForURL(event.itemPointer!)
                    }
                    if(item != nil && item!.isEqual(object)){
                        deleteEvent(event)
                    }
                }
            }
        }catch{
            print("error: \(error)")
        }
    }
    
    func getManagedObjectForURL(url: NSURL) -> NSManagedObject{
        let appDelegate = UIApplication.sharedApplication().delegate as! AppDelegate
        let objectId = appDelegate.getPersistentStoreCoordinator().managedObjectIDForURIRepresentation(url)
        return managedObjectContext!.objectWithID(objectId!)
    }

    func processQueue(){
        if(syncing == false){
            stop=false
            processFirstQueueEvent()
        }else{
            // Queue is already being processed.
            NSLog("queue is already being processed")
        }
    }
    
    func pause(){
        syncing = true // blocks new syncProcess requests
        stop = true
    }
    
    func resume(){
        stop = false
        syncing = false
    }
    
    func addSyncEventsRequest()->Event{
        // create an event without saving to DB
        let event = Event(entity: entity!, insertIntoManagedObjectContext: nil)
        event.method = HTTPMethod.POST
        event.handlerClass = "IncomingEventResolver"
        event.itemPointer = nil
        return event
    }
    
    func processFirstQueueEvent(){
        if(Settings.syncEnabled && Settings.username != ""){
            
            let fetchRequest = NSFetchRequest()
            fetchRequest.entity = NSEntityDescription.entityForName(Event.getEntityName(), inManagedObjectContext: managedObjectContext!)
            fetchRequest.sortDescriptors = [NSSortDescriptor(key: "time", ascending: true)]
            
            do{
                let results = try managedObjectContext!.executeFetchRequest(fetchRequest) as? [NSManagedObject]
                if(results!.count>0){
                    syncing = true
                    processEvent((results!.first) as! Event)
                }else if(checkServerUpdates){
                    // if all requests are completed and periodic timer fired, send a request for fetching updates from server
                    let event = addSyncEventsRequest()
                    processEvent(event)
                    checkServerUpdates = false
                }else{
                    syncing = false
                }
            }catch{
                print("error: \(error)")
            }
        }
    }
    
    func processEvent(event: Event){
        let handler = StorageFactory.createInstance(event.handlerClass, managedObjectContext: managedObjectContext!)
        let requestData = handler?.prepareRequestData(event)
        if(!stop){
            let request = ServerRequest(method: requestData!.event.method, url: requestData!.url, body: requestData!.body)
            request.run({ (request: NSURLRequest!, response: NSHTTPURLResponse!, data: AnyObject!) -> Void in
                //if(!self.stop){ // TODO, tohle by tu asi nemelo byt, server jiz odeslal odpoved a my ji musime zpracovat
                self.handleResponse(request, response: response, error: nil, data: data, requestData: requestData!, success: handler?.requestSuccessful)
                //}
            }, failure: { (request: NSURLRequest!, response: NSHTTPURLResponse!, error: NSError!, data: AnyObject!) -> Void in
                //if(!self.stop){ // TODO, tohle by tu asi nemelo byt, server jiz odeslal odpoved a my ji musime zpracovat
                self.handleResponse(request, response: response, error: error, data: data, requestData: requestData!, success: handler?.requestSuccessful)
                //}
            })
        }
    }
    
    func handleResponse(request: NSURLRequest!, response: NSHTTPURLResponse!, error: NSError?, data: AnyObject!, requestData: SyncRequestData, success: ((SyncRequestData!, NSHTTPURLResponse!, AnyObject?) -> Void)?){
        
        if(response.statusCode == BAD_LOGIN_CODE){ // bad login detection
            syncing = false
            badLogin()
        }else if(error == nil && response != nil){
            if(ACCEPTABLE_RESPONSE_CODES.contains(response.statusCode)){ // can be solved in central storage
                if (success != nil) {
                    success!(requestData, response, data)
                }
                self.deleteEvent(requestData.event)
                self.processFirstQueueEvent()
                hideErrorDialog = false
                hideBadLoginDialog = false
            }else{ // error code received
                syncing = false
                wrongResponse()
            }
        }else if(error != nil && error!.code == -1011 && error!.localizedRecoverySuggestion!.containsString("Device ID Lost")){ // device ID expired at the server, complete sync required
            performSyncBeginMerge()
            hideErrorDialog = false
            hideBadLoginDialog = false
        }else{ // internet connection offline (code -1009) or other error => stop processing the queue
            syncing = false
        }
    }
    
    func requestSucceedDump(request: NSURLRequest!, response: NSHTTPURLResponse!, data: AnyObject!){
        // test dump
        if data != nil {
            let json = JSON(data)
            NSLog("json - %@", json.rawString(0, options: NSJSONWritingOptions.PrettyPrinted)!)
        }
    }
    
    func badLogin(){
        if(hideBadLoginDialog == false){
            hideBadLoginDialog = true
            
            let msg = l("Bad login credentials detected! Please retype your username and password in settings.")
            let alert = UIAlertController(title: "Error", message: msg, preferredStyle: UIAlertControllerStyle.Alert)
            alert.addAction(UIAlertAction(title: "Hide", style: UIAlertActionStyle.Default, handler: { action in
                
            }))
            alert.addAction(UIAlertAction(title: "Settings", style: UIAlertActionStyle.Default, handler: { action in
                let settingsUrl = NSURL(string: UIApplicationOpenSettingsURLString)
                if let url = settingsUrl {
                    self.hideBadLoginDialog = true
                    UIApplication.sharedApplication().openURL(url)
                }
            }))
            showAlert(alert)
        }
    }
    
    /**
     Bad login dialog trigged by SyncBeginCtrlDelegate. After repairing login in settings sync begin is run again.
     */
    func badLogin(ctrl: SyncBeginController){
        if(SyncQueue.getInstance().hideBadLoginDialog == false){
            SyncQueue.getInstance().hideBadLoginDialog = true
            
            let msg = l("Bad login credentials detected! Please retype your username and password in settings.")
            let alert = UIAlertController(title: "Error", message: msg, preferredStyle: UIAlertControllerStyle.Alert)
            alert.addAction(UIAlertAction(title: "Hide", style: UIAlertActionStyle.Default, handler: { action in
                ctrl.dismissViewControllerAnimated(true, completion: nil)
            }))
            alert.addAction(UIAlertAction(title: "Settings", style: UIAlertActionStyle.Default, handler: { action in
                let settingsUrl = NSURL(string: UIApplicationOpenSettingsURLString)
                if let url = settingsUrl {
                    getAppDelegate().notifySyncBeginCtrl = true
                    getAppDelegate().syncBeginCtrl = ctrl
                    UIApplication.sharedApplication().openURL(url)
                }
            }))
            ctrl.presentViewController(alert, animated: true, completion: nil)
        }
    }
    
    func wrongResponse(){
        if(hideErrorDialog == false){
            hideErrorDialog = true
            
            let msg = l("We are sorry, but synchronisation failed! Probably it is a server error.")
            let alert = UIAlertController(title: "Error", message: msg, preferredStyle: UIAlertControllerStyle.Alert)
            alert.addAction(UIAlertAction(title: "Hide", style: UIAlertActionStyle.Default, handler: { action in
                self.hideErrorDialog = true
            }))
            alert.addAction(UIAlertAction(title: "Try again", style: UIAlertActionStyle.Default, handler: { action in
                self.hideErrorDialog = false
                self.processQueue()
            }))
            showAlert(alert)
        }
    }
    
    /**
    If the device ID expired, we need to run the complete synchronisation in the merge mode to prevent unsent events not to be deleted.
     */
    func performSyncBeginMerge(){
        let topCtrl = getTopViewController()
        let ctrl = topCtrl!.storyboard?.instantiateViewControllerWithIdentifier("sync-begin") as! UINavigationController
        let target = ctrl.viewControllers[0] as! SyncBeginController
        target.delegate = self
        target.useMerge = true
        topCtrl!.presentViewController(ctrl, animated: true, completion: nil)
    }
    
    /**
    Starts processing remaing events in the queue after SyncBegin merge is done (after expired device ID). 
    */
    func syncBeginSuccessful() {
        let topCtrl = getTopViewController()
        topCtrl?.dismissViewControllerAnimated(true, completion: nil)
        processQueue()
    }
}
